
from .model.VIPRS import VIPRS
from .model.VIPRSMix import VIPRSMix
from .model.gridsearch.VIPRSGridSearch import VIPRSGridSearch
from .model.gridsearch.HyperparameterGrid import HyperparameterGrid
from .utils.data_utils import *

__version__ = '0.1.1'
__release_date__ = 'April 2024'




from .binary_metrics import *
from .continuous_metrics import *

# Define a dictionary that maps evaluation metric names to their respective functions:
eval_metric_names = {
    'Pearson_R': pearson_r,
    'MSE': mse,
    'R2': r2,
    'Incremental_R2': incremental_r2,
    'Partial_Correlation': partial_correlation,
    'AUROC': roc_auc,
    'AUPRC': pr_auc,
    'Avg_Precision': avg_precision,
    'F1_Score': f1,
    'Liability_R2': liability_r2,
    'Liability_Probit_R2': liability_probit_r2,
    'Liability_Logit_R2': liability_logit_r2,
    'Nagelkerke_R2': nagelkerke_r2,
    'CoxSnell_R2': cox_snell_r2,
    'McFadden_R2': mcfadden_r2
}

# Define a list of metrics that can work with or require
# covariates to be computed:
eval_incremental_metrics = [
    'Incremental_R2',
    'Partial_Correlation',
    'Liability_R2',
    'Liability_Probit_R2',
    'Liability_Logit_R2',
    'Nagelkerke_R2',
    'CoxSnell_R2',
    'McFadden_R2'
]



# distutils: language = c++
# sources: model/vi/e_step.hpp

from cython cimport floating, integral


cdef extern from "e_step.hpp" nogil:

    bint blas_supported() noexcept nogil
    bint omp_supported() noexcept nogil

    void blas_axpy[T](T* y, T* x, T alpha, int size) noexcept nogil
    T blas_dot[T](T* x, T* y, int size) noexcept nogil

    void e_step[T, U, I](int c_size,
                      int* ld_left_bound,
                      I* ld_indptr,
                      U* ld_data,
                      T* std_beta,
                      T* var_gamma,
                      T* var_mu,
                      T* eta,
                      T* q,
                      T* eta_diff,
                      T* u_logs,
                      T* half_var_tau,
                      T* mu_mult,
                         T dq_scale,
                      int threads,
                      bint use_blas,
                      bint low_memory) noexcept nogil

    void e_step_mixture[T, U, I](int c_size,
                              int K,
                              int* ld_left_bound,
                              I* ld_indptr,
                              U* ld_data,
                              T* std_beta,
                              T* var_gamma,
                              T* var_mu,
                              T* eta,
                              T* q,
                              T* eta_diff,
                              T* log_null_pi,
                              T* u_logs,
                              T* half_var_tau,
                              T* mu_mult,
                                 T dq_scale,
                              int threads,
                              bint use_blas,
                              bint low_memory) noexcept nogil

    void e_step_grid[T, U, I](int c_size,
                           int n_active_models,
                           int* active_model_idx,
                           int* ld_left_bound,
                           I* ld_indptr,
                           U* ld_data,
                           T* std_beta,
                           T* var_gamma,
                           T* var_mu,
                           T* eta,
                           T* q,
                           T* eta_diff,
                           T* u_logs,
                           T* half_var_tau,
                           T* mu_mult,
                           T dq_scale,
                           int threads,
                           bint use_blas,
                           bint low_memory) noexcept nogil


cpdef check_blas_support():
    return blas_supported()


cpdef check_omp_support():
    return omp_supported()


cdef void cpp_blas_axpy(floating[::1] v1, floating[::1] v2, floating alpha) noexcept nogil:
    """v1 := v1 + alpha * v2"""
    cdef int size = v1.shape[0]
    blas_axpy(&v1[0], &v2[0], alpha, size)


cdef floating cpp_blas_dot(floating[::1] v1, floating[::1] v2) noexcept nogil:
    """v1 := v1.Tv2"""
    cdef int size = v1.shape[0]
    return blas_dot(&v1[0], &v2[0], size)


cpdef void cpp_e_step(int[::1] ld_left_bound,
                      integral[::1] ld_indptr,
                      noncomplex_numeric[::1] ld_data,
                      floating[::1] std_beta,
                      floating[::1] var_gamma,
                      floating[::1] var_mu,
                      floating[::1] eta,
                      floating[::1] q,
                      floating[::1] eta_diff,
                      floating[::1] u_logs,
                      floating[::1] half_var_tau,
                      floating[::1] mu_mult,
                      floating dq_scale,
                      int threads,
                      bint use_blas,
                      bint low_memory) noexcept nogil:

    e_step(var_mu.shape[0],
           &ld_left_bound[0],
           &ld_indptr[0],
           &ld_data[0],
           &std_beta[0],
           &var_gamma[0],
           &var_mu[0],
           &eta[0],
           &q[0],
           &eta_diff[0],
           &u_logs[0],
           &half_var_tau[0],
           &mu_mult[0],
           dq_scale,
           threads,
           use_blas,
           low_memory)


cpdef void cpp_e_step_mixture(int[::1] ld_left_bound,
                              integral[::1] ld_indptr,
                              noncomplex_numeric[::1] ld_data,
                              floating[::1] std_beta,
                              floating[:, ::1] var_gamma,
                              floating[:, ::1] var_mu,
                              floating[::1] eta,
                              floating[::1] q,
                              floating[::1] eta_diff,
                              floating[::1] log_null_pi,
                              floating[:, ::1] u_logs,
                              floating[:, ::1] half_var_tau,
                              floating[:, ::1] mu_mult,
                              floating dq_scale,
                              int threads,
                              bint use_blas,
                              bint low_memory) noexcept nogil:

    e_step_mixture(var_mu.shape[0],
                   var_mu.shape[1],
                   &ld_left_bound[0],
                   &ld_indptr[0],
                   &ld_data[0],
                   &std_beta[0],
                   &var_gamma[0, 0],
                   &var_mu[0, 0],
                   &eta[0],
                   &q[0],
                   &eta_diff[0],
                   &log_null_pi[0],
                   &u_logs[0, 0],
                   &half_var_tau[0, 0],
                   &mu_mult[0, 0],
                   dq_scale,
                   threads,
                   use_blas,
                   low_memory)

cpdef void cpp_e_step_grid(int[::1] ld_left_bound,
                           integral[::1] ld_indptr,
                           noncomplex_numeric[::1] ld_data,
                           floating[::1] std_beta,
                           floating[::1, :] var_gamma,
                           floating[::1, :] var_mu,
                           floating[::1, :] eta,
                           floating[::1, :] q,
                           floating[::1, :] eta_diff,
                           floating[::1, :] u_logs,
                           floating[::1, :] half_var_tau,
                           floating[::1, :] mu_mult,
                           floating dq_scale,
                           int[:] active_model_idx,
                           int threads,
                           bint use_blas,
                           bint low_memory) noexcept nogil:

    e_step_grid(var_mu.shape[0],
                active_model_idx.shape[0],
                &active_model_idx[0],
                &ld_left_bound[0],
                &ld_indptr[0],
                &ld_data[0],
                &std_beta[0],
                &var_gamma[0, 0],
                &var_mu[0, 0],
                &eta[0, 0],
                &q[0, 0],
                &eta_diff[0, 0],
                &u_logs[0, 0],
                &half_var_tau[0, 0],
                &mu_mult[0, 0],
                dq_scale,
                threads,
                use_blas,
                low_memory)



import numpy as np
from .VIPRSGrid import VIPRSGrid


class VIPRSGridSearch(VIPRSGrid):
    """
    The `VIPRSGridSearch` class is an extension of the `VIPRSGrid` class that
    implements grid search for the `VIPRS` models. The grid search procedure
    fits multiple models to the data, each with different hyperparameters,
    and selects the best model based on user-defined criteria.

    The criteria supported are:

    * `ELBO`: The model with the highest ELBO is selected.
    * `validation`: The model with the highest R^2 on the validation set is selected.
    * `pseudo_validation`: The model with the highest pseudo-validation R^2 is selected.

    Note that the `validation` and `pseudo_validation` criteria require the user to provide
    validation data in the form of paired genotype/phenotype data or external GWAS summary
    statistics.

    """

    def __init__(self,
                 gdl,
                 grid,
                 **kwargs):
        """
        Initialize the `VIPRSGridSearch` model.

        :param gdl: An instance of `GWADataLoader`
        :param grid: An instance of `HyperparameterGrid`
        :param kwargs: Additional keyword arguments to pass to the parent `VIPRSGrid` class.
        """

        super().__init__(gdl, grid=grid, **kwargs)

    def select_best_model(self, validation_gdl=None, criterion='ELBO'):
        """
        From the grid of models that were fit to the data, select the best 
        model according to the specified `criterion`. If the criterion is the ELBO,
        the model with the highest ELBO will be selected. If the criterion is
        validation or pseudo-validation, the model with the highest R^2 on the
        validation set will be selected.
        
        :param validation_gdl: An instance of `GWADataLoader` containing data from the validation set.
        Must be provided if criterion is `validation` or `pseudo_validation`.
        :param criterion: The criterion for selecting the best model. 
        Options are: (`ELBO`, `validation`, `pseudo_validation`)
        """

        assert criterion in ('ELBO', 'validation', 'pseudo_validation')

        if criterion == 'ELBO':
            best_model_idx = np.argmax(self.history['ELBO'][len(self.history['ELBO']) - 1])
        elif criterion == 'validation':

            assert validation_gdl is not None
            assert validation_gdl.sample_table is not None
            assert validation_gdl.sample_table.phenotype is not None

            from viprs.eval.continuous_metrics import r2

            prs = self.predict(test_gdl=validation_gdl)
            prs_r2 = [r2(prs[:, i], validation_gdl.sample_table.phenotype) for i in range(self.n_models)]
            self.validation_result['Validation_R2'] = prs_r2
            best_model_idx = np.argmax(prs_r2)
        elif criterion == 'pseudo_validation':

            pseudo_corr = self.pseudo_validate(validation_gdl, metric='r2')
            self.validation_result['Pseudo_Validation_Corr'] = pseudo_corr
            best_model_idx = np.argmax(np.nan_to_num(pseudo_corr, nan=-1., neginf=-1., posinf=-1.))

        if int(self.verbose) > 1:
            print(f"> Based on the {criterion} criterion, selected model: {best_model_idx}")
            print("> Model details:\n")
            print(self.validation_result.iloc[best_model_idx, :])

        # Update the variational parameters and their dependencies:
        for param in (self.pip, self.post_mean_beta, self.post_var_beta,
                      self.var_gamma, self.var_mu, self.var_tau,
                      self.eta, self.zeta, self.q):

            for c in param:
                param[c] = param[c][:, best_model_idx]

        # Update sigma epsilon:
        self.sigma_epsilon = self.sigma_epsilon[best_model_idx]

        # Update sigma beta:
        if isinstance(self.tau_beta, dict):
            for c in self.tau_beta:
                self.tau_beta[c] = self.tau_beta[c][:, best_model_idx]
        else:
            self.tau_beta = self.tau_beta[best_model_idx]

        # Update pi

        if isinstance(self.pi, dict):
            for c in self.pi:
                self.pi[c] = self.pi[c][:, best_model_idx]
        else:
            self.pi = self.pi[best_model_idx]

        # Set the number of models to 1:
        self.n_models = 1

        return self



import numpy as np
import pandas as pd
import os.path as osp

from ..utils.compute_utils import expand_column_names, dict_mean
from magenpy.utils.model_utils import merge_snp_tables


class BayesPRSModel:
    """
    A base class for Bayesian PRS models. This class defines the basic structure and methods
    that are common to most Bayesian PRS models. Specifically, this class provides methods and interfaces
    for initialization, harmonization, prediction, and fitting of Bayesian PRS models.

    The class is generic is designed to be inherited and extended by
    specific Bayesian PRS models, such as `LDPred` and `VIPRS`.

    :ivar gdl: A GWADataLoader object containing harmonized GWAS summary statistics and
    Linkage-Disequilibrium (LD) matrices.
    :ivar Nj: A dictionary where keys are chromosomes and values are the sample sizes per variant.
    :ivar shapes: A dictionary where keys are chromosomes and values are the shapes of the variant arrays
    (e.g. the number of variants per chromosome).
    :ivar _sample_size: The average per-SNP sample size.
    :ivar pip: The posterior inclusion probability.
    :ivar post_mean_beta: The posterior mean for the effect sizes.
    :ivar post_var_beta: The posterior variance for the effect sizes.
    """

    def __init__(self, gdl):
        """
        Initialize the Bayesian PRS model.
        :param gdl: An instance of `GWADataLoader`.
        """

        self.gdl = gdl

        # Sample size per SNP:
        try:
            self.Nj = {c: ss.n_per_snp.astype(float) for c, ss in gdl.sumstats_table.items()}
        except AttributeError:
            # If not provided, use the overall sample size:
            self.Nj = {c: np.repeat(gdl.n, c_size).astype(float) for c, c_size in gdl.shapes.items()}

        self.shapes = self.gdl.shapes.copy()

        # Determine the overall sample size:
        self._sample_size = dict_mean(self.Nj)

        # Inferred model parameters:
        self.pip = None  # Posterior inclusion probability
        self.post_mean_beta = None  # The posterior mean for the effect sizes
        self.post_var_beta = None  # The posterior variance for the effect sizes

    @property
    def chromosomes(self):
        """
        :return: The list of chromosomes that are included in the BayesPRSModel
        """
        return sorted(list(self.shapes.keys()))

    @property
    def m(self) -> int:
        """

        !!! seealso "See Also"
            * [n_snps][viprs.model.BayesPRSModel.BayesPRSModel.n_snps]

        :return: The number of variants in the model.
        """
        return self.gdl.m

    @property
    def n(self) -> int:
        """
        :return: The number of samples in the model. If not available, average the per-SNP
        sample sizes.
        """
        return self._sample_size

    @property
    def n_snps(self) -> int:
        """
        !!! seealso "See Also"
            * [m][viprs.model.BayesPRSModel.BayesPRSModel.m]

        :return: The number of SNPs in the model.
        """
        return self.m

    def fit(self, *args, **kwargs):
        """
        A genetic method to fit the Bayesian PRS model. This method should be implemented by the
        specific Bayesian PRS model.
        :raises NotImplementedError: If the method is not implemented in the child class.
        """
        raise NotImplementedError

    def get_proportion_causal(self):
        """
        A generic method to get an estimate of the proportion of causal variants.
        :raises NotImplementedError: If the method is not implemented in the child class.
        """
        raise NotImplementedError

    def get_heritability(self):
        """
        A generic method to get an estimate of the heritability, or proportion of variance explained by SNPs.
        :raises NotImplementedError: If the method is not implemented in the child class.
        """
        raise NotImplementedError

    def get_pip(self):
        """
        :return: The posterior inclusion probability for each variant in the model.
        """
        return self.pip

    def get_posterior_mean_beta(self):
        """
        :return: The posterior mean of the effect sizes (BETA) for each variant in the model.
        """
        return self.post_mean_beta

    def get_posterior_variance_beta(self):
        """
        :return: The posterior variance of the effect sizes (BETA) for each variant in the model.
        """
        return self.post_var_beta

    def predict(self, test_gdl=None):
        """
        Given the inferred effect sizes, predict the phenotype for the training samples in
        the GWADataLoader object or new test samples. If `test_gdl` is not provided, genotypes
        from training samples will be used (if available).

        :param test_gdl: A GWADataLoader object containing genotype data for new test samples.
        :raises ValueError: If the posterior means for BETA are not set. AssertionError if the GWADataLoader object
        does not contain genotype data.
        """

        if self.post_mean_beta is None:
            raise ValueError("The posterior means for BETA are not set. Call `.fit()` first.")

        if test_gdl is None:
            assert self.gdl.genotype is not None, "The GWADataLoader object must contain genotype data."
            test_gdl = self.gdl
            post_mean_beta = self.post_mean_beta
        else:
            _, post_mean_beta, _ = self.harmonize_data(gdl=test_gdl)

        return test_gdl.predict(post_mean_beta)

    def harmonize_data(self, gdl=None, parameter_table=None):
        """
        Harmonize the inferred effect sizes with a new GWADataLoader object. This method is useful
        when the user wants to predict on new samples or when the effect sizes are inferred from a
        different set of samples. The method aligns the effect sizes with the SNP table in the
        GWADataLoader object.

        :param gdl: An instance of `GWADataLoader` object.
        :param parameter_table: A `pandas` DataFrame of variant effect sizes.

        :return: A tuple of the harmonized posterior inclusion probability, posterior mean for the effect sizes,
        and posterior variance for the effect sizes.

        """

        if gdl is None and parameter_table is None:
            return

        if gdl is None:
            gdl = self.gdl

        if parameter_table is None:
            parameter_table = self.to_table(per_chromosome=True)
        else:
            parameter_table = {c: parameter_table.loc[parameter_table['CHR'] == c, ]
                               for c in parameter_table['CHR'].unique()}

        snp_tables = gdl.to_snp_table(col_subset=['SNP', 'A1', 'A2'],
                                      per_chromosome=True)

        pip = {}
        post_mean_beta = {}
        post_var_beta = {}

        common_chroms = sorted(list(set(snp_tables.keys()).intersection(set(parameter_table.keys()))))

        for c in common_chroms:

            try:
                post_mean_cols = expand_column_names('BETA', self.post_mean_beta[c].shape)
                if isinstance(post_mean_cols, str):
                    post_mean_cols = [post_mean_cols]

                pip_cols = expand_column_names('PIP', self.post_mean_beta[c].shape)
                if isinstance(pip_cols, str):
                    pip_cols = [pip_cols]

                post_var_cols = expand_column_names('VAR_BETA', self.post_mean_beta[c].shape)
                if isinstance(post_var_cols, str):
                    post_var_cols = [post_var_cols]

            except (TypeError, KeyError):
                pip_cols = [col for col in parameter_table[c].columns if 'PIP' in col]
                post_var_cols = [col for col in parameter_table[c].columns if 'VAR_BETA' in col]
                post_mean_cols = [col for col in parameter_table[c].columns
                                  if 'BETA' in col and col not in post_var_cols]

            # Merge the effect table with the GDL SNP table:
            c_df = merge_snp_tables(snp_tables[c], parameter_table[c], how='left',
                                    signed_statistics=post_mean_cols)

            if len(c_df) < len(snp_tables[c]):
                raise ValueError("The parameter table could not aligned with the reference SNP table. This may due to "
                                 "conflicts/errors in use of reference vs. alternative alleles.")

            # Obtain the values for the posterior mean:
            c_df[post_mean_cols] = c_df[post_mean_cols].fillna(0.)
            post_mean_beta[c] = c_df[post_mean_cols].values

            # Obtain the values for the posterior inclusion probability:
            if len(set(pip_cols).intersection(set(c_df.columns))) > 0:
                c_df[pip_cols] = c_df[pip_cols].fillna(0.)
                pip[c] = c_df[pip_cols].values

            # Obtain the values for the posterior variance:
            if len(set(post_var_cols).intersection(set(c_df.columns))) > 0:
                c_df[post_var_cols] = c_df[post_var_cols].fillna(0.)
                post_var_beta[c] = c_df[post_var_cols].values

        if len(pip) < 1:
            pip = None

        if len(post_var_beta) < 1:
            post_var_beta = None

        return pip, post_mean_beta, post_var_beta

    def to_table(self, col_subset=('CHR', 'SNP', 'A1', 'A2'), per_chromosome=False):
        """
        Output the posterior estimates for the effect sizes to a pandas dataframe.
        :param col_subset: The subset of columns to include in the tables (in addition to the effect sizes).
        :param per_chromosome: If True, return a separate table for each chromosome.

        :return: A pandas Dataframe with the posterior estimates for the effect sizes.
        """

        if self.post_mean_beta is None:
            raise Exception("The posterior means for BETA are not set. Call `.fit()` first.")

        tables = self.gdl.to_snp_table(col_subset=col_subset, per_chromosome=True)

        for c in self.chromosomes:

            tables[c][expand_column_names('BETA', self.post_mean_beta[c].shape)] = self.post_mean_beta[c]

            if self.pip is not None:
                tables[c][expand_column_names('PIP', self.pip[c].shape)] = self.pip[c]

            if self.post_var_beta is not None:
                tables[c][expand_column_names('VAR_BETA', self.post_var_beta[c].shape)] = self.post_var_beta[c]

        if per_chromosome:
            return tables
        else:
            return pd.concat([tables[c] for c in self.chromosomes])

    def pseudo_validate(self, test_gdl, metric='pearson_correlation'):
        """
        Evaluate the prediction accuracy of the inferred PRS using external GWAS summary statistics.

        :param test_gdl: A `GWADataLoader` object with the external GWAS summary statistics and LD matrix information.
        :param metric: The metric to use for evaluation. Options: 'r2' or 'pearson_correlation'.

        :return: The pseudo-validation metric.
        """

        from ..eval.pseudo_metrics import pseudo_r2, pseudo_pearson_r

        metric = metric.lower()

        assert self.post_mean_beta is not None, "The posterior means for BETA are not set. Call `.fit()` first."

        if metric in ('pearson_correlation', 'corr', 'r'):
            return pseudo_pearson_r(test_gdl, self.to_table(per_chromosome=False))
        elif metric == 'r2':
            return pseudo_r2(test_gdl, self.to_table(per_chromosome=False))
        else:
            raise KeyError(f"Pseudo validation metric ({metric}) not recognized. "
                           f"Options are: 'r2' or 'pearson_correlation'.")

    def set_model_parameters(self, parameter_table):
        """
        Parses a pandas dataframe with model parameters and assigns them 
        to the corresponding class attributes. 
        
        For example: 
            * Columns with `BETA`, will be assigned to `self.post_mean_beta`.
            * Columns with `PIP` will be assigned to `self.pip`.
            * Columns with `VAR_BETA`, will be assigned to `self.post_var_beta`.
        
        :param parameter_table: A pandas table or dataframe.
        """

        self.pip, self.post_mean_beta, self.post_var_beta = self.harmonize_data(parameter_table=parameter_table)

    def read_inferred_parameters(self, f_names, sep=r"\s+"):
        """
        Read a file with the inferred parameters.
        :param f_names: A path (or list of paths) to the file with the effect sizes.
        :param sep: The delimiter for the file(s).
        """

        if isinstance(f_names, str):
            f_names = [f_names]

        param_table = []

        for f_name in f_names:
            param_table.append(pd.read_csv(f_name, sep=sep))

        if len(param_table) > 0:
            param_table = pd.concat(param_table)
            self.set_model_parameters(param_table)
        else:
            raise FileNotFoundError

    def write_inferred_parameters(self, f_name, per_chromosome=False, sep="\t"):
        """
        A convenience method to write the inferred posterior for the effect sizes to file.
        :param f_name: The filename (or directory) where to write the effect sizes
        :param per_chromosome: If True, write a file for each chromosome separately.
        :param sep: The delimiter for the file (tab by default).
        """

        tables = self.to_table(per_chromosome=per_chromosome)

        if '.fit' not in f_name:
            ext = '.fit'
        else:
            ext = ''

        if per_chromosome:
            for c, tab in tables.items():
                try:
                    tab.to_csv(osp.join(f_name, f'chr_{c}.fit'), sep=sep, index=False)
                except Exception as e:
                    raise e
        else:
            try:
                tables.to_csv(f_name + ext, sep=sep, index=False)
            except Exception as e:
                raise e



# cython: linetrace=False
# cython: profile=False
# cython: binding=False
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: nonecheck=False
# cython: language_level=3
# cython: infer_types=True

cimport cython
import numpy as np
from cython cimport floating
from libc.math cimport exp, log
from scipy.linalg.cython_blas cimport saxpy, daxpy, sdot, ddot


# ------------------------------------------------------------
# BLAS implementations of some linear algebra operations

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.exceptval(check=False)
cdef void scipy_blas_axpy(floating[::1] v1, floating[::1] v2, floating alpha) noexcept nogil:
    """v1 := v1 + alpha * v2"""
    cdef:
        int inc = 1, n=v1.shape[0]

    if floating is float:
        saxpy(&n, &alpha, &v2[0], &inc, &v1[0], &inc)
    else:
        daxpy(&n, &alpha, &v2[0], &inc, &v1[0], &inc)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.exceptval(check=False)
cdef floating scipy_blas_dot(floating[::1] v1, floating[::1] v2) noexcept nogil:
    """v1 . v2"""
    cdef:
        int inc = 1, n=v1.shape[0]

    if floating is float:
        return sdot(&n, &v1[0], &inc, &v2[0], &inc)
    else:
        return ddot(&n, &v1[0], &inc, &v2[0], &inc)

# ------------------------------------------------------------

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
@cython.exceptval(check=False)
cdef floating[::1] softmax(floating[::1] x) noexcept nogil:
    """
    A numerically stable implementation of softmax
    """

    cdef unsigned int i, end = x.shape[0]
    cdef floating s = 0., max_x = c_max(x)

    with nogil:
        for i in range(end):
            x[i] = exp(x[i] - max_x)
            s += x[i]

        for i in range(end):
            x[i] /= s

    return x

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
@cython.exceptval(check=False)
cdef floating sigmoid(floating x) noexcept nogil:
    """
    A numerically stable version of the Sigmoid function.
    """
    if x < 0:
        exp_x = exp(x)
        return exp_x / (1. + exp_x)
    else:
        return 1. / (1. + exp(-x))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
@cython.exceptval(check=False)
cdef floating logit(floating x) noexcept nogil:
    """
    The logit function (inverse of the sigmoid function)
    """
    return log(x / (1. - x))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.exceptval(check=False)
cdef floating dot(floating[::1] v1, floating[::1] v2) noexcept nogil:
    """
    Dot product between vectors of the same shape
    """

    cdef unsigned int i, end = v1.shape[0]
    cdef floating s = 0.

    with nogil:
        for i in range(end):
            s += v1[i]*v2[i]

    return s

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.exceptval(check=False)
cdef floating vec_sum(floating[::1] v1) noexcept nogil:
    """
    Vector summation
    """

    cdef unsigned int i, end = v1.shape[0]
    cdef floating s = 0.

    with nogil:
        for i in range(end):
            s += v1[i]

    return s

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.exceptval(check=False)
cdef void axpy(floating[::1] v1, floating[::1] v2, floating s) noexcept nogil:
    """
    Elementwise addition and multiplication
    """

    cdef unsigned int i, end = v1.shape[0]

    with nogil:
        for i in range(end):
            v1[i] = v1[i] + v2[i] * s

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
@cython.exceptval(check=False)
cdef floating[::1] clip_list(floating[::1] a, floating min_value, floating max_value) noexcept nogil:
    """
    Iterate over a list and clip every element to be between `min_value` and `max_value`
    :param a: A list of floating point numbers
    :param min_value: Minimum values
    :param max_value: Maximum value
    """

    cdef unsigned int i, end = a.shape[0]

    with nogil:
        for i in range(end):
            a[i] = clip(a[i], min_value, max_value)

    return a

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
@cython.exceptval(check=False)
cdef floating c_max(floating[::1] x) noexcept nogil:
    """
    Obtain the maximum value in a vector `x`
    """
    cdef unsigned int i, end = x.shape[0]
    cdef floating current_max = 0.

    with nogil:
        for i in range(end):
            if i == 0 or current_max < x[i]:
                current_max = x[i]

    return current_max

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.exceptval(check=False)
cdef floating clip(floating a, floating min_value, floating max_value) noexcept nogil:
    """
    Clip a scalar value `a` to be between `min_value` and `max_value`
    """

    if a < min_value:
        a = min_value
    if a > max_value:
        a = max_value

    return a

def bernoulli_entropy(p):
    """
    Compute the entropy of a Bernoulli variable given a vector of probabilities.
    :param p: A vector (or scalar) of probabilities between zero and one, 0. < p < 1.
    """
    return -(p*np.log(p) + (1. - p)*np.log(1. - p))



from magenpy import GWADataLoader
import numpy as np


def _match_variant_stats(test_gdl, prs_beta_table):
    """
    Match the standardized marginal betas from the validation set to the inferred PRS effect sizes.
    This function takes a `GWADataLoader` object from the validation set (with matched LD matrices and
    GWAS summary statistics) and a PRS table object and returns a tuple of three arrays:

    #. The standardized marginal betas from the validation set
    #. The inferred PRS effect sizes
    #. The LD-weighted PRS effect sizes (q)

    :param test_gdl: A `GWADataLoader` object from the validation or test set.
    :param prs_beta_table: A pandas DataFrame with the PRS effect sizes. Must contain
    the columns: CHR, SNP, A1, A2, BETA.

    :return: A tuple of three arrays: (1) The standardized marginal betas from the validation set,
    (2) The inferred PRS effect sizes, (3) The LD-weighted PRS effect sizes (q).
    """

    # Sanity checks:
    assert isinstance(test_gdl, GWADataLoader), "The test/validation set must be an instance of GWADataLoader."
    assert test_gdl.ld is not None, "The test/validation set must have LD matrices initialized."
    assert test_gdl.sumstats_table is not None, "The test/validation set must have summary statistics initialized."

    from magenpy.utils.model_utils import merge_snp_tables

    validation_tab = test_gdl.to_snp_table(col_subset=['CHR', 'SNP', 'A1', 'A2', 'STD_BETA'],
                                           per_chromosome=True)

    required_cols = ['CHR', 'SNP', 'A1', 'A2']
    for col in required_cols:
        assert col in prs_beta_table.columns, f"The PRS effect sizes table must contain a column named {col}."

    validation_beta = []
    prs_beta = []
    ld_weighted_beta = []

    if 'BETA' in prs_beta_table.columns:
        beta_cols = ['BETA']
    else:
        beta_cols = [col for col in prs_beta_table.columns if 'BETA' in col and 'VAR' not in col]
        assert len(beta_cols) > 0, "The PRS effect sizes table must contain a column named BETA or BETA_0, BETA_1, etc."

    per_chrom_prs_tables = dict(tuple(prs_beta_table.groupby('CHR')))

    for chrom, tab in validation_tab.items():

        if chrom not in per_chrom_prs_tables:
            continue

        c_df = merge_snp_tables(tab,
                                per_chrom_prs_tables[chrom],
                                how='left',
                                signed_statistics=beta_cols)

        test_gdl.ld[chrom].load(dtype=np.float32)

        validation_beta.append(tab['STD_BETA'].values)
        prs_beta.append(c_df[beta_cols].fillna(0.).values)
        ld_weighted_beta.append(test_gdl.ld[chrom].dot(prs_beta[-1]))

        test_gdl.ld[chrom].release()

    return (np.concatenate(validation_beta),
            np.concatenate(prs_beta, axis=0),
            np.concatenate(ld_weighted_beta))


def pseudo_r2(test_gdl, prs_beta_table):
    """
    Compute the R-Squared metric (proportion of variance explained) for a given
    PRS using standardized marginal betas from an independent test set.
    Here, we follow the pseudo-validation procedures outlined in Mak et al. (2017) and
    Yang and Zhou (2020), where the proportion of phenotypic variance explained by the PRS
    in an independent validation cohort can be approximated with:

    R2(PRS, y) ~= 2*r'b - b'Sb

    Where `r` is the standardized marginal beta from a validation/test set,
    `b` is the posterior mean for the effect size of each variant and `S` is the LD matrix.

    :param test_gdl: An instance of `GWADataLoader` with the summary statistics table initialized.
    :param prs_beta_table: A pandas DataFrame with the PRS effect sizes. Must contain
    the columns: CHR, SNP, A1, A2, BETA.
    """

    std_beta, prs_beta, q = _match_variant_stats(test_gdl, prs_beta_table)

    rb = np.sum((prs_beta.T * std_beta).T, axis=0)
    bsb = np.sum(prs_beta*q, axis=0)

    return 2*rb - bsb


def pseudo_pearson_r(test_gdl, prs_beta_table):
    """
    Perform pseudo-validation of the inferred effect sizes by comparing them to
    standardized marginal betas from an independent validation set. Here, we follow the pseudo-validation
    procedures outlined in Mak et al. (2017) and Yang and Zhou (2020), where
    the correlation between the PRS and the phenotype in an independent validation
    cohort can be approximated with:

    Corr(PRS, y) ~= r'b / sqrt(b'Sb)

    Where `r` is the standardized marginal beta from a validation set,
    `b` is the posterior mean for the effect size of each variant and `S` is the LD matrix.

    :param test_gdl: An instance of `GWADataLoader` with the summary statistics table initialized.
    :param prs_beta_table: A pandas DataFrame with the PRS effect sizes. Must contain
    the columns: CHR, SNP, A1, A2, BETA.
    """

    std_beta, prs_beta, q = _match_variant_stats(test_gdl, prs_beta_table)

    rb = np.sum((prs_beta.T * std_beta).T, axis=0)
    bsb = np.sum(prs_beta * q, axis=0)

    return rb / np.sqrt(bsb)




import numpy as np
from sklearn.metrics import (
    auc,
    roc_auc_score,
    average_precision_score,
    precision_recall_curve,
    f1_score
)
from .continuous_metrics import incremental_r2
from scipy.stats import norm
import statsmodels.api as sm
import pandas as pd


def roc_auc(true_val, pred_val):
    """
    Compute the area under the ROC (AUROC) for a model
     that maps from the PRS predictions to the binary phenotype.

    :param true_val: The response value or phenotype (a numpy binary vector with 0s and 1s)
    :param pred_val: The predicted value or PRS (a numpy vector)
    """
    return roc_auc_score(true_val, pred_val)


def pr_auc(true_val, pred_val):
    """
    Compute the area under the Precision-Recall curve for a model
    that maps from the PRS predictions to the binary phenotype.

    :param true_val: The response value or phenotype (a binary numpy vector with 0s and 1s)
    :param pred_val: The predicted value or PRS (a numpy vector)
    """
    precision, recall, thresholds = precision_recall_curve(true_val, pred_val)
    return auc(recall, precision)


def avg_precision(true_val, pred_val):
    """
    Compute the average precision between the PRS predictions and a binary.

    :param true_val: The response value or phenotype (a binary numpy vector with 0s and 1s)
    :param pred_val: The predicted value or PRS (a numpy vector)
    """
    return average_precision_score(true_val, pred_val)


def f1(true_val, pred_val):
    """
    Compute the F1 score between the PRS predictions and a binary phenotype.

    :param true_val: The response value or phenotype (a binary numpy vector with 0s and 1s)
    :param pred_val: The predicted value or PRS (a numpy vector)
    """
    return f1_score(true_val, pred_val)


def mcfadden_r2(true_val, pred_val, covariates=None):
    """
    Compute the McFadden pseudo-R^2 between the PRS predictions and a phenotype.
    If covariates are provided, we compute the incremental pseudo-R^2 by conditioning
    on the covariates.


    :param true_val: The response value or phenotype (a binary numpy vector with 0s and 1s)
    :param pred_val: The predicted value or PRS (a numpy vector)
    :param covariates: A pandas table of covariates where the rows are ordered
    the same way as the predictions and response.
    """

    if covariates is None:
        covariates = pd.DataFrame(np.ones((true_val.shape[0], 1)), columns=['const'])
    else:
        covariates = sm.add_constant(covariates)

    null_result = sm.Logit(true_val, covariates).fit(disp=0)
    full_result = sm.Logit(true_val, covariates.assign(pred_val=pred_val)).fit(disp=0)

    return 1. - (full_result.llf / null_result.llf)


def cox_snell_r2(true_val, pred_val, covariates=None):
    """
    Compute the Cox-Snell pseudo-R^2 between the PRS predictions and a binary phenotype.
    If covariates are provided, we compute the incremental pseudo-R^2 by conditioning
    on the covariates.

    :param true_val: The response value or phenotype (a binary numpy vector with 0s and 1s)
    :param pred_val: The predicted value or PRS (a numpy vector)
    :param covariates: A pandas table of covariates where the rows are ordered
    the same way as the predictions and response.
    """

    if covariates is None:
        covariates = pd.DataFrame(np.ones((true_val.shape[0], 1)), columns=['const'])
    else:
        covariates = sm.add_constant(covariates)

    null_result = sm.Logit(true_val, covariates).fit(disp=0)
    full_result = sm.Logit(true_val, covariates.assign(pred_val=pred_val)).fit(disp=0)
    n = true_val.shape[0]

    return 1. - np.exp(-2 * (full_result.llf - null_result.llf) / n)


def nagelkerke_r2(true_val, pred_val, covariates=None):
    """
    Compute the Nagelkerke pseudo-R^2 between the PRS predictions and a binary phenotype.
    If covariates are provided, we compute the incremental pseudo-R^2 by conditioning
    on the covariates.

    :param true_val: The response value or phenotype (a binary numpy vector with 0s and 1s)
    :param pred_val: The predicted value or PRS (a numpy vector)
    :param covariates: A pandas table of covariates where the rows are ordered
    the same way as the predictions and response.
    """

    if covariates is None:
        covariates = pd.DataFrame(np.ones((true_val.shape[0], 1)), columns=['const'])
    else:
        covariates = sm.add_constant(covariates)

    null_result = sm.Logit(true_val, covariates).fit(disp=0)
    full_result = sm.Logit(true_val, covariates.assign(pred_val=pred_val)).fit(disp=0)
    n = true_val.shape[0]

    # First compute the Cox & Snell R2:
    cox_snell = 1. - np.exp(-2 * (full_result.llf - null_result.llf) / n)

    # Then scale it by the maximum possible R2:
    return cox_snell / (1. - np.exp(2 * null_result.llf / n))


def liability_r2(true_val, pred_val, covariates=None, return_all_r2=False):
    """
    Compute the coefficient of determination (R^2) on the liability scale
    according to Lee et al. (2012) Gene. Epi.
    https://pubmed.ncbi.nlm.nih.gov/22714935/

    The R^2 liability is defined as:
    R_{liability}^2 = R2_{observed}*K*(K-1)/(z^2)

    where R_{observed}^2 is the R^2 on the observed scale and K is the sample prevalence
    and z is the "height of the normal density at the quantile for K".

    If covariates are provided, we compute the incremental pseudo-R^2 by conditioning
    on the covariates.

    :param true_val: The response value or phenotype (a binary numpy vector with 0s and 1s)
    :param pred_val: The predicted value or PRS (a numpy vector)
    :param covariates: A pandas table of covariates where the rows are ordered
    the same way as the predictions and response.
    :param return_all_r2: If True, return the null, full and incremental R2 values.
    """

    # First, obtain the incremental R2 on the observed scale:
    r2_obs = incremental_r2(true_val, pred_val, covariates, return_all_r2=return_all_r2)

    # Second, compute the prevalence and the standard normal quantile of the prevalence:

    k = np.mean(true_val)
    z2 = norm.pdf(norm.ppf(1.-k))**2
    mult_factor = k*(1. - k) / z2

    if return_all_r2:
        return {
            'Null_R2': r2_obs['Null_R2']*mult_factor,
            'Full_R2': r2_obs['Full_R2']*mult_factor,
            'Incremental_R2': r2_obs['Incremental_R2']*mult_factor
        }
    else:
        return r2_obs * mult_factor


def liability_probit_r2(true_val, pred_val, covariates=None, return_all_r2=False):
    """
    Compute the R^2 between the PRS predictions and a binary phenotype on the liability
    scale using the probit likelihood as outlined in Lee et al. (2012) Gene. Epi.
    https://pubmed.ncbi.nlm.nih.gov/22714935/

    The R^2 is defined as:
    R2_{probit} = Var(pred) / (Var(pred) + 1)

    Where Var(pred) is the variance of the predicted liability.

    If covariates are provided, we compute the incremental pseudo-R^2 by conditioning
    on the covariates.

    :param true_val: The response value or phenotype (a binary numpy vector with 0s and 1s)
    :param pred_val: The predicted value or PRS (a numpy vector)
    :param covariates: A pandas table of covariates where the rows are ordered
    the same way as the predictions and response.
    :param return_all_r2: If True, return the null, full and incremental R2 values.
    """

    if covariates is None:
        covariates = pd.DataFrame(np.ones((true_val.shape[0], 1)), columns=['const'])
    else:
        covariates = sm.add_constant(covariates)

    null_result = sm.Probit(true_val, covariates).fit(disp=0)
    full_result = sm.Probit(true_val, covariates.assign(pred_val=pred_val)).fit(disp=0)

    null_var = np.var(null_result.predict())
    null_r2 = null_var / (null_var + 1.)

    full_var = np.var(full_result.predict())
    full_r2 = full_var / (full_var + 1.)

    if return_all_r2:
        return {
            'Null_R2': null_r2,
            'Full_R2': full_r2,
            'Incremental_R2': full_r2 - null_r2
        }
    else:
        return full_r2 - null_r2


def liability_logit_r2(true_val, pred_val, covariates=None, return_all_r2=False):
    """
    Compute the R^2 between the PRS predictions and a binary phenotype on the liability
    scale using the logit likelihood as outlined in Lee et al. (2012) Gene. Epi.
    https://pubmed.ncbi.nlm.nih.gov/22714935/

    The R^2 is defined as:
    R2_{probit} = Var(pred) / (Var(pred) + pi^2 / 3)

    Where Var(pred) is the variance of the predicted liability.

    If covariates are provided, we compute the incremental pseudo-R^2 by conditioning
    on the covariates.

    :param true_val: The response value or phenotype (a binary numpy vector with 0s and 1s)
    :param pred_val: The predicted value or PRS (a numpy vector)
    :param covariates: A pandas table of covariates where the rows are ordered
    the same way as the predictions and response.
    :param return_all_r2: If True, return the null, full and incremental R2 values.
    """

    if covariates is None:
        covariates = pd.DataFrame(np.ones((true_val.shape[0], 1)), columns=['const'])
    else:
        covariates = sm.add_constant(covariates)

    null_result = sm.Probit(true_val, covariates).fit(disp=0)
    full_result = sm.Probit(true_val, covariates.assign(pred_val=pred_val)).fit(disp=0)

    null_var = np.var(null_result.predict())
    null_r2 = null_var / (null_var + (np.pi**2 / 3))

    full_var = np.var(full_result.predict())
    full_r2 = full_var / (full_var + (np.pi**2 / 3))

    if return_all_r2:
        return {
            'Null_R2': null_r2,
            'Full_R2': full_r2,
            'Incremental_R2': full_r2 - null_r2
        }
    else:
        return full_r2 - null_r2



import numpy as np
import logging
from tqdm import tqdm

from .BayesPRSModel import BayesPRSModel
from magenpy.stats.h2.ldsc import simple_ldsc
from ..utils.exceptions import OptimizationDivergence
from .vi.e_step import e_step
from .vi.e_step_cpp import cpp_e_step
from ..utils.OptimizeResult import OptimizeResult
from ..utils.compute_utils import dict_mean, dict_sum, dict_concat


class VIPRS(BayesPRSModel):
    """
    The base class for performing Variational Inference of Polygenic Risk Scores (VIPRS).

    This class implements the Variational EM algorithm for estimating the posterior distribution
    of the effect sizes using GWAS summary statistics. The model assumes a spike-and-slab mixture
    prior on the effect size distribution, with the spike component representing the null effects
    and the slab component representing the non-null effects.

    Details for the algorithm can be found in the Supplementary Material of the following paper:

    > Zabad S, Gravel S, Li Y. Fast and accurate Bayesian polygenic risk modeling with variational inference.
    Am J Hum Genet. 2023 May 4;110(5):741-761. doi: 10.1016/j.ajhg.2023.03.009.
    Epub 2023 Apr 7. PMID: 37030289; PMCID: PMC10183379.

    :ivar gdl: An instance of GWADataLoader containing harmonized GWAS summary statistics and LD matrices.
    :ivar var_gamma: A dictionary of the variational gamma parameter, denoting the probability that the
    variant comes from the slab component.
    :ivar var_mu: A dictionary of the variational mu parameter, denoting the mean of the
    effect size for each variant.
    :ivar var_tau: A dictionary of the variational tau parameter, denoting the precision of
    the effect size for each variant.
    :ivar eta: A dictionary of the posterior mean of the effect size, E[B] = gamma*mu.
    :ivar zeta: A dictionary of the expectation of B^2 under the posterior, E[B^2] = gamma*(mu^2 + 1./tau).
    :ivar eta_diff: A dictionary of the difference between the etas in two consecutive iterations.
    :ivar q: A dictionary of the q-factor, which keeps track of the multiplication of eta with the LD matrix.
    :ivar ld_data: A dictionary of the `data` arrays of the sparse LD matrices.
    :ivar ld_indptr: A dictionary of the `indptr` arrays of the sparse LD matrices.
    :ivar ld_left_bound: A dictionary of the left boundaries of the LD matrices.
    :ivar std_beta: A dictionary of the standardized marginal effect sizes from GWAS.
    :ivar Nj: A dictionary of the sample size per SNP from the GWAS study.
    :ivar threads: The number of threads to use when fitting the model.
    :ivar fix_params: A dictionary of hyperparameters with their fixed values.
    :ivar float_precision: The precision of the floating point variables. Options are: 'float32' or 'float64'.
    :ivar order: The order of the arrays in memory. Options are: 'C' or 'F'.
    :ivar low_memory: A boolean flag to indicate whether to use low memory mode.
    :ivar dequantize_on_the_fly: A boolean flag to indicate whether to dequantize the LD matrix on the fly.
    :ivar use_cpp: A boolean flag to indicate whether to use the C++ backend.
    :ivar use_blas: A boolean flag to indicate whether to use BLAS for linear algebra operations.
    :ivar optim_result: An instance of OptimizeResult tracking the progress of the optimization algorithm.
    :ivar verbose: Verbosity of the information printed to standard output. Can be boolean or an integer.
    :ivar history: A dictionary to store the history of the optimization procedure (e.g. the objective as a function
    of iteration number).
    :ivar tracked_theta: A list of hyperparameters to track throughout the optimization procedure. Useful for
    debugging/model checking.

    """

    def __init__(self,
                 gdl,
                 fix_params=None,
                 tracked_theta=None,
                 verbose=True,
                 float_precision='float32',
                 order='F',
                 low_memory=False,
                 use_blas=True,
                 use_cpp=True,
                 dequantize_on_the_fly=False,
                 threads=1):

        """

        Initialize the VIPRS model.

        .. note::
            The initialization of the model involves loading the LD matrix to memory.

        :param gdl: An instance of GWADataLoader containing harmonized GWAS summary statistics and LD matrices.
        :param fix_params: A dictionary of hyperparameters with their fixed values.
        :param tracked_theta: A list of hyperparameters to track throughout the optimization procedure. Useful
        for debugging/model checking. Currently, we allow the user to track the following:

            * The proportion of causal variants (`pi`).
            * The heritability ('heritability').
            * The residual variance (`sigma_epsilon`).

        :param verbose: Verbosity of the information printed to standard output. Can be boolean or an integer.
        Provide a number greater than 1 for more detailed output.
        :param float_precision: The precision of the floating point variables. Options are: 'float32' or 'float64'.
        :param order: The order of the arrays in memory. Options are: 'C' or 'F'.
        :param low_memory: A boolean flag to indicate whether to use low memory mode.
        :param use_blas: A boolean flag to indicate whether to use BLAS for linear algebra operations.
        :param use_cpp: A boolean flag to indicate whether to use the C++ backend.
        :param dequantize_on_the_fly: A boolean flag to indicate whether to dequantize the LD matrix on the fly.
        :param threads: The number of threads to use when fitting the model.
        """

        super().__init__(gdl)

        # ------------------- Sanity checks -------------------

        assert gdl.ld is not None, "The LD matrices must be initialized in the GWADataLoader object."
        assert gdl.sumstats_table is not None, ("The summary statistics must be "
                                                "initialized in the GWADataLoader object.")

        if dequantize_on_the_fly and not use_cpp:
            raise Exception("Dequantization on the fly is only supported when using the C++ backend.")

        # ------------------- Initialize the model -------------------

        # Variational parameters:
        self.var_gamma = {}
        self.var_mu = {}
        self.var_tau = {}

        # Properties of proposed variational distribution:
        self.eta = {}  # The posterior mean, E[B] = \gamma*\mu_beta
        self.zeta = {}  # The expectation of B^2 under the posterior, E[B^2] = \gamma*(\mu_beta^2 + 1./\tau_beta)

        # The difference between the etas in two consecutive iterations (can be used for checking convergence,
        # or implementing optimized updates in the E-Step).
        self.eta_diff = {}

        # q-factor (keeps track of LD-related terms)
        self.q = {}

        # ---------- Inputs to the model: ----------

        # NOTE: Here, we typecast the inputs to the model to the specified float precision.
        # This also needs to be done in the initialization methods.

        # LD-related quantities:

        self.ld_data = {}
        self.ld_indptr = {}
        self.ld_left_bound = {}

        for c, ld_mat in self.gdl.get_ld_matrices().items():
            # Load the data for the LD matrix:
            if dequantize_on_the_fly and np.issubdtype(ld_mat.stored_dtype, np.integer):
                # Cannot dequantize float16 on the fly due to lack of canonical representation
                # for this data type:
                assert ld_mat.stored_dtype in [np.int8, np.int16, np.int32, np.int64]
                dtype = ld_mat.stored_dtype
            else:

                if dequantize_on_the_fly:
                    print("Warning: Dequantization on the fly is only supported for "
                          "integer data types. Ignoring this flag.")

                dtype = float_precision
                dequantize_on_the_fly = False

            if low_memory:

                self.ld_data[c], self.ld_indptr[c] = ld_mat.low_memory_load(dtype=dtype)
                self.ld_left_bound[c] = np.arange(1, len(self.ld_indptr[c]) - 1, dtype=np.int32)
            else:
                ld_mat.load(return_symmetric=True, fill_diag=True, dtype=dtype)

                self.ld_data[c] = ld_mat.data
                self.ld_indptr[c] = ld_mat.indptr
                self.ld_left_bound[c] = ld_mat.ld_boundaries[0, :]

        # Standardized betas:
        self.std_beta = {c: ss.get_snp_pseudo_corr().astype(float_precision)
                         for c, ss in self.gdl.sumstats_table.items()}

        # Make sure that the data type for the sample size-per-SNP has the correct format:

        self.Nj = {c: nj.astype(float_precision, order=order)
                   for c, nj in self.Nj.items()}

        # ---------- General properties: ----------

        self.threads = threads
        self.fix_params = fix_params or {}

        self.float_precision = float_precision
        self.float_resolution = np.finfo(self.float_precision).resolution
        self.order = order
        self.low_memory = low_memory

        self.dequantize_on_the_fly = dequantize_on_the_fly

        if self.dequantize_on_the_fly:
            info = np.iinfo(ld_mat.dtype)
            self.dequantize_scale = 2. / (info.max - (info.min + 1))
        else:
            self.dequantize_scale = 1.

        self.use_cpp = use_cpp
        self.use_blas = use_blas

        self.optim_result = OptimizeResult()
        self.verbose = verbose
        self.history = {}
        self.tracked_theta = tracked_theta or []

    def initialize(self, theta_0=None, param_0=None):
        """
        A convenience method to initialize all the objects associated with the model.
        :param theta_0: A dictionary of initial values for the hyperparameters theta
        :param param_0: A dictionary of initial values for the variational parameters
        """

        if int(self.verbose) > 1:
            print("> Initializing model parameters")

        self.initialize_theta(theta_0)
        self.initialize_variational_parameters(param_0)
        self.init_history()
        self.optim_result.reset()

    def init_history(self):
        """
        Initialize the history object to track various quantities of interest, such as
        the optimization objective (ELBO). This method should be called before fitting the model.
        """

        self.history = {
            'ELBO': [],
            'old ELBO': [],
        }

        for tt in self.tracked_theta:
            self.history[tt] = []

    def initialize_theta(self, theta_0=None):
        """
        Initialize the global hyperparameters of the model.
        :param theta_0: A dictionary of initial values for the hyperparameters theta
        """

        if theta_0 is not None and self.fix_params is not None:
            theta_0.update(self.fix_params)
        elif self.fix_params is not None:
            theta_0 = self.fix_params
        elif theta_0 is None:
            theta_0 = {}

        # ----------------------------------------------
        # (1) If 'pi' is not set, initialize from a uniform
        if 'pi' not in theta_0:
            self.pi = np.random.uniform(low=max(0.005, 1. / self.n_snps), high=.1)
        else:
            self.pi = theta_0['pi']

        # ----------------------------------------------
        # (2) Initialize sigma_epsilon and tau_beta
        # Assuming that the genotype and phenotype are normalized,
        # these two quantities are conceptually linked.
        # The initialization routine here assumes that:
        # Var(y) = h2 + sigma_epsilon
        # Where, by assumption, Var(y) = 1,
        # And h2 ~= pi*M/tau_beta

        if 'sigma_epsilon' not in theta_0:
            if 'tau_beta' not in theta_0:

                # If neither tau_beta nor sigma_epsilon are given,
                # then initialize using the SNP heritability estimate

                try:
                    naive_h2g = np.clip(simple_ldsc(self.gdl), a_min=1e-3, a_max=1. - 1e-3)
                except Exception as e:
                    naive_h2g = np.random.uniform(low=.001, high=.999)

                self.sigma_epsilon = 1. - naive_h2g
                self.tau_beta = self.pi * self.n_snps / naive_h2g
            else:

                # If tau_beta is given, use it to initialize sigma_epsilon

                self.tau_beta = theta_0['tau_beta']
                self.sigma_epsilon = np.clip(1. - (self.pi * self.n_snps / self.tau_beta),
                                             a_min=self.float_resolution,
                                             a_max=1. - self.float_resolution)
        else:

            # If sigma_epsilon is given, use it in the initialization

            self.sigma_epsilon = theta_0['sigma_epsilon']

            if 'tau_beta' in theta_0:
                self.tau_beta = theta_0['tau_beta']
            else:
                self.tau_beta = (self.pi * self.n_snps) / (1. - self.sigma_epsilon)

        # Cast all the hyperparameters to conform to the precision set by the user:
        self.sigma_epsilon = np.dtype(self.float_precision).type(self.sigma_epsilon)
        self.tau_beta = np.dtype(self.float_precision).type(self.tau_beta)
        self.pi = np.dtype(self.float_precision).type(self.pi)

    def initialize_variational_parameters(self, param_0=None):
        """
        Initialize the variational parameters of the model.
        :param param_0: A dictionary of initial values for the variational parameters
        """

        param_0 = param_0 or {}

        self.var_mu = {}
        self.var_tau = {}
        self.var_gamma = {}

        for c, shapes in self.shapes.items():

            # Initialize the variational parameters according to the derived update equations,
            # ignoring correlations between SNPs.
            if 'tau' in param_0:
                self.var_tau[c] = param_0['tau'][c]
            else:
                self.var_tau[c] = (self.Nj[c] / self.sigma_epsilon) + self.tau_beta

            if 'mu' in param_0:
                self.var_mu[c] = param_0['mu'][c]
            else:
                self.var_mu[c] = np.zeros(shapes)

            if 'gamma' in param_0:
                self.var_gamma[c] = param_0['gamma'][c]
            else:
                pi = self.get_pi(c)
                if isinstance(self.pi, dict):
                    self.var_gamma[c] = pi.copy()
                else:
                    self.var_gamma[c] = pi*np.ones(shapes)

            # Ensure that all the variational parameters are set to the desired
            # float precision:
            self.var_mu[c] = self.var_mu[c].astype(self.float_precision, order=self.order)
            self.var_tau[c] = self.var_tau[c].astype(self.float_precision, order=self.order)
            self.var_gamma[c] = self.var_gamma[c].astype(self.float_precision, order=self.order)

        self.eta = self.compute_eta()
        self.zeta = self.compute_zeta()
        self.eta_diff = {c: np.zeros_like(eta, dtype=self.float_precision) for c, eta in self.eta.items()}
        self.q = {c: np.zeros_like(eta, dtype=self.float_precision) for c, eta in self.eta.items()}

    def e_step(self):
        """
        Run the E-Step of the Variational EM algorithm.
        Here, we update the variational parameters for each variant using coordinate
        ascent optimization techniques. The update equations are outlined in
        the Supplementary Material of the following paper:

        > Zabad S, Gravel S, Li Y. Fast and accurate Bayesian polygenic risk modeling with variational inference.
        Am J Hum Genet. 2023 May 4;110(5):741-761. doi: 10.1016/j.ajhg.2023.03.009.
        Epub 2023 Apr 7. PMID: 37030289; PMCID: PMC10183379.
        """

        for c, c_size in self.shapes.items():

            # Get the priors:
            tau_beta = self.get_tau_beta(c)
            pi = self.get_pi(c)

            # Updates for tau variational parameters:
            self.var_tau[c] = (self.Nj[c] / self.sigma_epsilon) + tau_beta

            # Compute some quantities that are needed for the per-SNP updates:
            mu_mult = self.Nj[c]/(self.var_tau[c]*self.sigma_epsilon)
            u_logs = np.log(pi) - np.log(1. - pi) + .5*(np.log(tau_beta) - np.log(self.var_tau[c]))

            if self.use_cpp:
                cpp_e_step(self.ld_left_bound[c],
                           self.ld_indptr[c],
                           self.ld_data[c],
                           self.std_beta[c],
                           self.var_gamma[c],
                           self.var_mu[c],
                           self.eta[c],
                           self.q[c],
                           self.eta_diff[c],
                           u_logs,
                           0.5*self.var_tau[c],
                           mu_mult,
                           self.dequantize_scale,
                           self.threads,
                           self.use_blas,
                           self.low_memory)
            else:

                e_step(self.ld_left_bound[c],
                       self.ld_indptr[c],
                       self.ld_data[c],
                       self.std_beta[c],
                       self.var_gamma[c],
                       self.var_mu[c],
                       self.eta[c],
                       self.q[c],
                       self.eta_diff[c],
                       u_logs,
                       0.5*self.var_tau[c],
                       mu_mult,
                       self.threads,
                       self.use_blas,
                       self.low_memory)

        self.zeta = self.compute_zeta()

    def update_pi(self):
        """
        Update the prior probability of a variant being causal, or the proportion of causal variants, `pi`.
        """

        if 'pi' not in self.fix_params:

            # Get the average of the gammas:
            self.pi = dict_mean(self.var_gamma, axis=0)

    def update_tau_beta(self):
        """
        Update the prior precision (inverse variance) for the effect size, `tau_beta`.
        """

        if 'tau_beta' not in self.fix_params:

            # tau_beta estimate:
            self.tau_beta = (self.pi * self.m / dict_sum(self.zeta, axis=0)).astype(self.float_precision)

    def update_sigma_epsilon(self):
        """
        Update the global residual variance parameter, `sigma_epsilon`.
        """

        if 'sigma_epsilon' not in self.fix_params:

            sig_eps = 0.

            for c, _ in self.shapes.items():

                sig_eps += np.sum(
                    - 2.*np.multiply(self.eta[c].T, self.std_beta[c]).T +
                    self.zeta[c] +
                    np.multiply(self.eta[c], self.q[c])
                , axis=0)

            self.sigma_epsilon = 1. + sig_eps

    def m_step(self):
        """
        Run the M-Step of the Variational EM algorithm.
        Here, we update the hyperparameters of the model, by simply calling
        the update functions for each hyperparameter separately.

        """

        self.update_pi()
        self.update_tau_beta()
        self.update_sigma_epsilon()

    def objective(self):
        """
        The optimization objective for the variational inference problem. The objective
        for the VIPRS method is the Evidence Lower-Bound (ELBO) in this case.

        !!! seealso "See Also"
            * [elbo][viprs.model.VIPRS.VIPRS.elbo]


        """
        return self.elbo()

    def elbo(self, sum_axis=None):
        """
        Compute the variational objective, the Evidence Lower-BOund (ELBO),
        from GWAS summary statistics and the reference LD data. This implementation assumes
        that the product of the LD matrix with the current estimate of the effect sizes
        is already computed and stored in the `q` dictionary. If this is not the case,
        we recommend computing q first and then calling this method.

        :param sum_axis: The axis along which to sum the ELBO. If None, the ELBO is returned as a scalar.
        """

        # Concatenate the dictionary items for easy computation:
        var_gamma = np.clip(dict_concat(self.var_gamma),
                            a_min=self.float_resolution,
                            a_max=1. - self.float_resolution)
        # The gamma for the null component
        null_gamma = np.clip(1. - dict_concat(self.compute_pip()),
                             a_min=self.float_resolution,
                             a_max=1. - self.float_resolution)
        var_mu = dict_concat(self.var_mu)
        var_tau = dict_concat(self.var_tau)

        if isinstance(self.pi, dict):
            pi = dict_concat(self.pi)
            null_pi = dict_concat(self.get_null_pi())
        else:
            pi = self.pi
            null_pi = self.get_null_pi()

        if isinstance(self.tau_beta, dict):
            tau_beta = dict_concat(self.tau_beta)
        else:
            tau_beta = self.tau_beta

        q = dict_concat(self.q)
        eta = dict_concat(self.eta)
        zeta = dict_concat(self.zeta)

        std_beta = dict_concat(self.std_beta)

        elbo = 0.

        # -----------------------------------------------
        # (1) Compute the log of the joint density:

        #
        # (1.1) The following terms are an expansion of ||Y - X\beta||^2
        #
        # -N/2log(2pi*sigma_epsilon)
        elbo -= np.log(2 * np.pi * self.sigma_epsilon)

        # -Y'Y/(2*sigma_epsilon), where we assume Y'Y = N
        # + (1./sigma_epsilon)*\beta*(XY), where we assume XY = N\hat{\beta}
        elbo -= (1. / self.sigma_epsilon) * (1. - 2.*np.sum(np.multiply(eta.T, std_beta).T, axis=0) +
                                             np.sum(np.multiply(eta, q) + zeta, axis=0))

        elbo *= 0.5*self.n

        elbo -= np.multiply(var_gamma, np.log(var_gamma) - np.log(pi)).sum(axis=sum_axis)
        elbo -= np.multiply(null_gamma, np.log(null_gamma) - np.log(null_pi)).sum(axis=sum_axis)

        elbo += .5 * np.multiply(var_gamma, 1. - np.log(var_tau) + np.log(tau_beta) -
                                 tau_beta*var_mu**2 - tau_beta/var_tau).sum(axis=sum_axis)

        try:
            if len(elbo) == 1:
                return elbo[0]
            else:
                return elbo
        except TypeError:
            return elbo

    def get_sigma_epsilon(self):
        """
        :return: The value of the residual variance, `sigma_epsilon`.
        """
        return self.sigma_epsilon

    def get_tau_beta(self, chrom=None):
        """
        :param chrom: Get the value of `tau_beta` for a given chromosome.

        :return: The value of the prior precision on the effect size(s), `tau_beta`
        """
        if chrom is None:
            return self.tau_beta
        else:
            if isinstance(self.tau_beta, dict):
                return self.tau_beta[chrom]
            else:
                return self.tau_beta

    def get_pi(self, chrom=None):
        """
        :param chrom: Get the value of `pi` for a given chromosome.

        :return: The value of the prior probability of a variant being causal, `pi`.
        """

        if chrom is None:
            return self.pi
        else:
            if isinstance(self.pi, dict):
                return self.pi[chrom]
            else:
                return self.pi

    def get_null_pi(self, chrom=None):
        """
        :param chrom: If provided, get the mixing proportion for the null component on a given chromosome.

        :return: The value of the prior probability of a variant being null, `1 - pi`.
        """

        pi = self.get_pi(chrom=chrom)

        if isinstance(pi, dict):
            return {c: 1. - c_pi for c, c_pi in pi.items()}
        else:
            return 1. - pi

    def get_proportion_causal(self):
        """
        :return: The proportion of causal variants in the model.
        """
        if isinstance(self.pi, dict):
            return dict_mean(self.pi, axis=0)
        else:
            return self.pi

    def get_average_effect_size_variance(self):
        """
        :return: The average per-SNP variance for the prior mixture components
        """
        if isinstance(self.pi, dict):
            pi = dict_concat(self.pi, axis=0)
        else:
            pi = self.pi

        if isinstance(self.tau_beta, dict):
            tau_beta = dict_concat(self.tau_beta, axis=0)
        else:
            tau_beta = self.tau_beta

        return np.sum(pi / tau_beta, axis=0)

    def get_heritability(self):
        """
        :return: An estimate of the SNP heritability, or proportion of variance explained by SNPs.
        """

        sigma_g = np.sum([
            np.sum(self.zeta[c] + np.multiply(self.q[c], self.eta[c]), axis=0)
            for c in self.shapes.keys()
        ], axis=0)

        h2g = sigma_g / (sigma_g + self.sigma_epsilon)

        return h2g

    def to_theta_table(self):
        """
        :return: A `pandas` DataFrame containing information about the estimated hyperparameters of the model.
        """

        theta_table = [
            {'Parameter': 'Residual_variance', 'Value': self.sigma_epsilon},
            {'Parameter': 'Heritability', 'Value': self.get_heritability()},
            {'Parameter': 'Proportion_causal', 'Value': self.get_proportion_causal()},
            {'Parameter': 'Average_effect_variance', 'Value': self.get_average_effect_size_variance()}
        ]

        if isinstance(self.tau_beta, dict):
            taus = dict_mean(self.tau_beta, axis=0)
        else:
            taus = self.tau_beta

        try:
            taus = list(taus)
            for i in range(len(taus)):
                theta_table.append({'Parameter': f'tau_beta_{i+1}', 'Value': taus[i]})
        except TypeError:
            theta_table.append({'Parameter': 'tau_beta', 'Value': taus})

        import pandas as pd

        return pd.DataFrame(theta_table)

    def write_inferred_theta(self, f_name, sep="\t"):
        """
        A convenience method to write the inferred (and fixed) hyperparameters of the model to file.
        :param f_name: The file name
        :param sep: The separator for the hyperparameter file.
        """

        # Write the table to file:
        try:
            self.to_theta_table().to_csv(f_name, sep=sep, index=False)
        except Exception as e:
            raise e

    def update_theta_history(self):
        """
        A convenience method to update the history of the hyperparameters of the model,
        if the user requested that they should be tracked.
        """

        for tt in self.tracked_theta:
            if tt == 'pi':
                self.history['pi'].append(self.get_proportion_causal())
            elif tt == 'pis':
                self.history['pis'].append(self.pi)
            if tt == 'heritability':
                self.history['heritability'].append(self.get_heritability())
            if tt == 'sigma_epsilon':
                self.history['sigma_epsilon'].append(self.sigma_epsilon)
            elif tt == 'tau_beta':
                self.history['tau_beta'].append(self.tau_beta)

    def compute_pip(self):
        """
        :return: The posterior inclusion probability
        """
        return self.var_gamma.copy()

    def compute_eta(self):
        """
        :return: The mean for the effect size under the variational posterior.
        """
        return {c: v*self.var_mu[c] for c, v in self.var_gamma.items()}

    def compute_zeta(self):
        """
        :return: The expectation of the squared effect size under the variational posterior.
        """
        return {c: np.multiply(v, self.var_mu[c]**2 + 1./self.var_tau[c])
                for c, v in self.var_gamma.items()}

    def update_posterior_moments(self):
        """
        A convenience method to update the dictionaries containing the posterior moments,
        including the PIP and posterior mean and variance for the effect size.
        """

        self.pip = {c: pip.copy() for c, pip in self.compute_pip().items()}
        self.post_mean_beta = {c: eta.copy() for c, eta in self.eta.items()}
        self.post_var_beta = {c: zeta - self.eta[c]**2 for c, zeta in self.zeta.items()}

    def fit(self,
            max_iter=1000,
            theta_0=None,
            param_0=None,
            continued=False,
            f_abs_tol=1e-6,
            x_abs_tol=1e-6,
            drop_r_tol=0.01,
            patience=5):
        """
        A convenience method to fit the model using the Variational EM algorithm.

        :param max_iter: Maximum number of iterations. 
        :param theta_0: A dictionary of values to initialize the hyperparameters
        :param param_0: A dictionary of values to initialize the variational parameters
        :param continued: If true, continue the model fitting for more iterations from current parameters
        instead of starting over.
        :param f_abs_tol: The absolute tolerance threshold for the objective (ELBO).
        :param x_abs_tol: The absolute tolerance threshold for the variational parameters.
        :param drop_r_tol: The relative tolerance for the drop in the ELBO to be considered as a red flag. It usually
        happens around convergence that the objective fluctuates due to numerical errors. This is a way to
        differentiate such random fluctuations from actual drops in the objective.
        :param patience: The maximum number of times the objective is allowed to drop before termination.
        """

        if not continued:
            self.initialize(theta_0, param_0)
            start_idx = 1
        else:
            start_idx = len(self.history['ELBO']) + 1

        if int(self.verbose) > 1:
            print("> Performing model fit...")
            if self.threads > 1:
                print(f"> Using up to {self.threads} threads.")

        # If the model is fit over a single chromosome, append this information to the
        # tqdm progress bar:
        if len(self.shapes) == 1:
            desc = f"Chromosome {list(self.shapes.keys())[0]}"
        else:
            desc = None

        # Progress bar:
        pbar = tqdm(range(start_idx, start_idx + max_iter),
                    disable=not self.verbose,
                    desc=desc)

        for i in pbar:

            if self.optim_result.stop_iteration:
                pbar.set_postfix({'Final ELBO': f"{self.optim_result.objective:.4f}"})
                pbar.n = i - 1
                pbar.total = i - 1
                pbar.refresh()
                pbar.close()
                break

            self.update_theta_history()

            self.e_step()
            self.m_step()

            self.history['ELBO'].append(self.elbo())
            pbar.set_postfix({'ELBO': f"{self.history['ELBO'][-1]:.4f}"})

            if i > 1:

                curr_elbo = self.history['ELBO'][-1]
                prev_elbo = self.history['ELBO'][-2]

                # Check for convergence in the objective + parameters:
                if np.isclose(prev_elbo, curr_elbo, atol=f_abs_tol, rtol=0.):
                    self.optim_result.update(curr_elbo,
                                             stop_iteration=True,
                                             success=True,
                                             message='Objective (ELBO) converged successfully.')

                # Check for convergence based on eta_diff?

                # Check to see if the objective drops due to numerical instabilities:
                elif curr_elbo < prev_elbo and not np.isclose(curr_elbo, prev_elbo, atol=0., rtol=drop_r_tol):
                    patience -= 1

                    if patience == 0:
                        self.optim_result.update(curr_elbo,
                                                 stop_iteration=True,
                                                 success=False,
                                                 message='Optimization is halted due to numerical instabilities.')
                    else:
                        self.optim_result.update(curr_elbo)

                    # Continue so as not to update the posterior moments in this case
                    continue

                # Check if the model parameters behave in unexpected/pathological ways:
                elif np.isnan(curr_elbo):
                    raise OptimizationDivergence(f"Stopping at iteration {i}: "
                                                 f"The optimization algorithm is not converging!\n"
                                                 f"The objective (ELBO) is NaN.")
                elif self.sigma_epsilon <= 0.:
                    raise OptimizationDivergence(f"Stopping at iteration {i}: "
                                                 f"The optimization algorithm is not converging!\n"
                                                 f"The residual variance estimate is negative.")
                elif self.get_heritability() >= 1.:
                    raise OptimizationDivergence(f"Stopping at iteration {i}: "
                                                 f"The optimization algorithm is not converging!\n"
                                                 f"Value of estimated heritability exceeded 1.")
                else:
                    self.optim_result.update(curr_elbo)

            self.update_posterior_moments()

        if not self.optim_result.stop_iteration:
            self.optim_result.update(self.history['ELBO'][-1],
                                     stop_iteration=True,
                                     success=False,
                                     message="Maximum iterations reached without convergence.\n"
                                             "You may need to run the model for more iterations.")

        if self.optim_result.success:
            if int(self.verbose) > 1:
                print(f"Converged at iteration {self.optim_result.iterations} | "
                      f"ELBO: {self.optim_result.objective:.6f}")
        else:
            logging.warning("\t" + self.optim_result.message)

        if int(self.verbose) > 1:
            print(f"> Final ELBO: {self.history['ELBO'][-1]:.6f}")
            print(f"> Estimated heritability: {self.get_heritability():.6f}")
            print(f"> Estimated proportion of causal variants: {self.get_proportion_causal():.6f}")

        return self



import pandas as pd
import numpy as np

from magenpy.stats.h2.ldsc import simple_ldsc
from .VIPRS import VIPRS
from .vi.e_step import e_step_mixture
from .vi.e_step_cpp import cpp_e_step_mixture
from ..utils.compute_utils import dict_sum, dict_mean


class VIPRSMix(VIPRS):
    """
    A class for the Variational Inference for Polygenic Risk Scores (VIPRS) model
    parametrized with the sparse mixture prior on the effect sizes. The class inherits
    many of the methods and attributes from the `VIPRS` class unchanged. However,
    there are many important updates and changes to the model, including the dimensionality
    of the arrays representing the variational parameters.

    Details for the algorithm can be found in the Supplementary Material of the following paper:

    > Zabad S, Gravel S, Li Y. Fast and accurate Bayesian polygenic risk modeling with variational inference.
    Am J Hum Genet. 2023 May 4;110(5):741-761. doi: 10.1016/j.ajhg.2023.03.009.
    Epub 2023 Apr 7. PMID: 37030289; PMCID: PMC10183379.

    :ivar K: The number of causal (i.e. non-null) components in the mixture prior (minimum 1). When `K=1`, this
    effectively reduces `VIPRSMix` to the `VIPRS` model.
    :ivar d: Multiplier for the prior on the effect size (vector of size K).

    """

    def __init__(self,
                 gdl,
                 K=1,
                 prior_multipliers=None,
                 **kwargs):

        """
        :param gdl: An instance of `GWADataLoader`
        :param K: The number of causal (i.e. non-null) components in the mixture prior (minimum 1). When `K=1`, this
            effectively reduces `VIPRSMix` to the `VIPRS` model.
        :param prior_multipliers: Multiplier for the prior on the effect size (vector of size K).
        :param kwargs: Additional keyword arguments to pass to the VIPRS model.
        """

        # Make sure that the matrices follow the C-contiguous order:
        kwargs['order'] = 'C'

        super().__init__(gdl, **kwargs)

        # Sanity checks:
        assert K > 0  # Check that there is at least 1 causal component
        self.K = K

        if prior_multipliers is not None:
            assert len(prior_multipliers) == K
            self.d = np.array(prior_multipliers).astype(self.float_precision)
        else:
            self.d = 2**np.linspace(-min(K - 1, 7), 0, K).astype(self.float_precision)

        # Populate/update relevant fields:
        self.shapes = {c: (shp, self.K) for c, shp in self.shapes.items()}
        self.Nj = {c: Nj[:, None].astype(self.float_precision, order=self.order) for c, Nj in self.Nj.items()}

    def initialize_theta(self, theta_0=None):
        """
        Initialize the global hyperparameters of the model
        :param theta_0: A dictionary of initial values for the hyperparameters theta
        """

        if theta_0 is not None and self.fix_params is not None:
            theta_0.update(self.fix_params)
        elif self.fix_params is not None:
            theta_0 = self.fix_params
        elif theta_0 is None:
            theta_0 = {}

        # ----------------------------------------------
        # (1) Initialize pi from a uniform
        if 'pis' in theta_0:
            self.pi = theta_0['pis']
        else:
            if 'pi' in theta_0:
                overall_pi = theta_0['pi']
            else:
                overall_pi = np.random.uniform(low=max(0.005, 1. / self.n_snps), high=.1)

            self.pi = overall_pi*np.random.dirichlet(np.ones(self.K))

        # ----------------------------------------------
        # (2) Initialize sigma_epsilon and sigma_beta
        # Assuming that the genotype and phenotype are normalized,
        # these two quantities are conceptually linked.
        # The initialization routine here assumes that:
        # Var(y) = h2 + sigma_epsilon
        # Where, by assumption, Var(y) = 1,
        # And h2 ~= pi*M*sigma_beta

        if 'sigma_epsilon' not in theta_0:

            if 'tau_betas' in theta_0:

                # If tau_betas are given, use them to initialize sigma_epsilon

                self.tau_beta = theta_0['tau_betas']

                self.sigma_epsilon = np.clip(1. - np.dot(1./self.tau_beta, self.pi),
                                             a_min=self.float_resolution,
                                             a_max=1. - self.float_resolution)

            elif 'tau_beta' in theta_0:
                # NOTE: Here, we assume the provided `tau_beta` is a scalar.
                # This is different from `tau_betas`

                assert self.d is not None

                self.tau_beta = theta_0['tau_beta'] * self.d
                # Use the provided tau_beta to initialize sigma_epsilon.
                # First, we derive a naive estimate of the heritability, based on the following equation:
                # h2g/M = \sum_k pi_k \tau_k
                # Where the per-SNP heritability is defined by the sum over the mixtures.

                # Step (1): Given the provided tau_beta and associated multipliers,
                # obtain a naive estimate of the heritability:
                h2g_estimate = (self.n_snps*self.pi/self.tau_beta).sum()
                # Step (2): Set sigma_epsilon to 1 - h2g_estimate:
                self.sigma_epsilon = np.clip(1. - h2g_estimate,
                                             a_min=self.float_resolution,
                                             a_max=1. - self.float_resolution)

            else:
                # If neither sigma_beta nor sigma_epsilon are given,
                # then initialize using the SNP heritability estimate based on summary statistics

                try:
                    naive_h2g = np.clip(simple_ldsc(self.gdl), 1e-3, 1. - 1e-3)
                except Exception as e:
                    naive_h2g = np.random.uniform(low=.001, high=.999)

                self.sigma_epsilon = 1. - naive_h2g

                global_tau = (self.n_snps * np.dot(1./self.d, self.pi) / naive_h2g)

                self.tau_beta = self.d*global_tau
        else:

            # If sigma_epsilon is given, use it in the initialization

            self.sigma_epsilon = theta_0['sigma_epsilon']

            # Initialize tau_betas
            if 'tau_betas' in theta_0:
                self.tau_beta = theta_0['tau_betas']
            elif 'tau_beta' in theta_0:
                self.tau_beta = np.repeat(theta_0['tau_beta'], self.K)
            else:
                # If not provided, initialize using sigma_epsilon value
                global_tau = (self.n_snps * np.dot(1./self.d, self.pi) / (1. - self.sigma_epsilon))

                self.tau_beta = self.d * global_tau

        # Cast all the hyperparameters to conform to the precision set by the user:
        self.sigma_epsilon = np.dtype(self.float_precision).type(self.sigma_epsilon)
        self.tau_beta = np.dtype(self.float_precision).type(self.tau_beta)
        self.pi = np.dtype(self.float_precision).type(self.pi)

    def e_step(self):
        """
        Run the E-Step of the Variational EM algorithm.
        Here, we update the variational parameters for each variant using coordinate
        ascent optimization techniques. The update equations are outlined in
        the Supplementary Material of the following paper:

        > Zabad S, Gravel S, Li Y. Fast and accurate Bayesian polygenic risk modeling with variational inference.
        Am J Hum Genet. 2023 May 4;110(5):741-761. doi: 10.1016/j.ajhg.2023.03.009.
        Epub 2023 Apr 7. PMID: 37030289; PMCID: PMC10183379.
        """

        for c, shapes in self.shapes.items():

            # Get the priors:
            tau_beta = self.get_tau_beta(c)
            pi = self.get_pi(c)

            # Updates for tau variational parameters:
            self.var_tau[c] = (self.Nj[c] / self.sigma_epsilon) + tau_beta

            if isinstance(self.pi, dict):
                log_null_pi = (np.log(1. - self.pi[c].sum(axis=1)))
            else:
                log_null_pi = np.ones_like(self.eta[c])*np.log(1. - self.pi.sum())

            # Compute some quantities that are needed for the per-SNP updates:
            mu_mult = self.Nj[c] / (self.var_tau[c] * self.sigma_epsilon)
            u_logs = np.log(pi) - np.log(1. - pi) + .5 * (np.log(tau_beta) - np.log(self.var_tau[c]))

            if self.use_cpp:
                cpp_e_step_mixture(self.ld_left_bound[c],
                                   self.ld_indptr[c],
                                   self.ld_data[c],
                                   self.std_beta[c],
                                   self.var_gamma[c],
                                   self.var_mu[c],
                                   self.eta[c],
                                   self.q[c],
                                   self.eta_diff[c],
                                   log_null_pi,
                                   u_logs,
                                   0.5*self.var_tau[c],
                                   mu_mult,
                                   self.dequantize_scale,
                                   self.threads,
                                   self.use_blas,
                                   self.low_memory)
            else:
                e_step_mixture(self.ld_left_bound[c],
                               self.ld_indptr[c],
                               self.ld_data[c],
                               self.std_beta[c],
                               self.var_gamma[c],
                               self.var_mu[c],
                               self.eta[c],
                               self.q[c],
                               self.eta_diff[c],
                               log_null_pi,
                               u_logs,
                               0.5*self.var_tau[c],
                               mu_mult,
                               self.threads,
                               self.use_blas,
                               self.low_memory)

        self.zeta = self.compute_zeta()

    def update_pi(self):
        """
        Update the prior mixing proportions `pi`
        """

        if 'pis' not in self.fix_params:

            pi_estimate = dict_sum(self.var_gamma, axis=0)

            if 'pi' in self.fix_params:
                # If the user provides an estimate for the total proportion of causal variants,
                # update the pis such that the proportion of SNPs in the null component becomes 1. - pi.
                pi_estimate = self.fix_params['pi']*pi_estimate / pi_estimate.sum()
            else:
                pi_estimate /= self.n_snps

            # Set pi to the new estimate:
            self.pi = pi_estimate

    def update_tau_beta(self):
        """
        Update the prior precision (inverse variance) for the effect sizes, `tau_beta`
        """

        if 'tau_betas' not in self.fix_params:

            # If a list of multipliers is provided,
            # estimate the global sigma_beta and then multiply it
            # by the per-component multiplier to get the final sigma_betas.

            zetas = sum(self.compute_zeta(sum_axis=0).values())

            tau_beta_estimate = np.sum(self.pi)*self.m / np.dot(self.d, zetas)
            tau_beta_estimate = self.d*tau_beta_estimate

            self.tau_beta = np.clip(tau_beta_estimate, a_min=1., a_max=None)

    def get_null_pi(self, chrom=None):
        """
        Get the proportion of SNPs in the null component
        :param chrom: If provided, get the mixing proportion for the null component on a given chromosome.
        :return: The value of the mixing proportion for the null component
        """

        pi = self.get_pi(chrom=chrom)

        if isinstance(pi, dict):
            return {c: 1. - c_pi.sum(axis=1) for c, c_pi in pi.items()}
        else:
            return 1. - np.sum(pi)

    def get_proportion_causal(self):
        """
        :return: The proportion of variants in the non-null components.
        """
        if isinstance(self.pi, dict):
            dict_mean({c: pis.sum(axis=1) for c, pis in self.pi.items()})
        else:
            return np.sum(self.pi)

    def get_average_effect_size_variance(self):
        """
        :return: The average per-SNP variance for the prior mixture components
        """

        avg_sigma = super().get_average_effect_size_variance()

        try:
            return avg_sigma.sum()
        except Exception:
            return avg_sigma

    def compute_pip(self):
        """
        :return: The posterior inclusion probability
        """
        return {c: gamma.sum(axis=1) for c, gamma in self.var_gamma.items()}

    def compute_eta(self):
        """
        :return: The mean for the effect size under the variational posterior.
        """
        return {c: (v * self.var_mu[c]).sum(axis=1) for c, v in self.var_gamma.items()}

    def compute_zeta(self, sum_axis=1):
        """
        :return: The expectation of the squared effect size under the variational posterior.
        """
        return {c: (v * (self.var_mu[c] ** 2 + (1./self.var_tau[c]))).sum(axis=sum_axis)
                for c, v in self.var_gamma.items()}

    def to_theta_table(self):
        """
        :return: A `pandas` DataFrame containing information about the estimated hyperparameters of the model.
        """

        table = super().to_theta_table()

        extra_theta = []

        if isinstance(self.pi, dict):
            pis = list(dict_mean(self.pi, axis=0))
        else:
            pis = self.pi

        for i in range(self.K):
            extra_theta.append({'Parameter': f'pi_{i + 1}', 'Value': pis[i]})

        return pd.concat([table, pd.DataFrame(extra_theta)])




def download_ukb_wb_ld_matrix(target_dir='.', chromosome=None):
    """
    Download the LD matrix for the White British samples in the UK Biobank.
    :param target_dir: The path or directory where to store the LD matrix
    :param chromosome: An integer or list of integers with the chromosome numbers for which to download
    the LD matrices from Zenodo.
    """

    import urllib.request
    from magenpy.utils.system_utils import makedir
    from magenpy.utils.compute_utils import iterable
    import os
    from tqdm import tqdm

    if chromosome is None:
        chromosome = list(range(1, 23))
    elif not iterable(chromosome):
        chromosome = [chromosome]

    if len(chromosome) < 2:
        print("> Download LD matrix for chromosome", chromosome[0])

    for c in tqdm(chromosome, total=len(chromosome), disable=len(chromosome) < 2,
                  desc='Downloading LD matrices'):

        target_path = os.path.join(target_dir, f"chr_{c}.tar.gz")

        try:
            makedir(os.path.dirname(target_path))
            urllib.request.urlretrieve(f"https://zenodo.org/record/7036625/files/chr_{c}.tar.gz?download=1",
                                       target_path)
            os.system(f"tar -xvzf {target_path}")
        except Exception as e:
            print(e)



from cython cimport floating, integral
cimport numpy as np

ctypedef fused noncomplex_numeric:
    np.int8_t
    np.int16_t
    np.int32_t
    np.int64_t
    np.float32_t
    np.float64_t


cdef void cpp_blas_axpy(floating[::1] v1, floating[::1] v2, floating alpha) noexcept nogil
cdef floating cpp_blas_dot(floating[::1] v1, floating[::1] v2) noexcept nogil


cpdef void cpp_e_step(int[::1] ld_left_bound,
                      integral[::1] ld_indptr,
                      noncomplex_numeric[::1] ld_data,
                      floating[::1] std_beta,
                      floating[::1] var_gamma,
                      floating[::1] var_mu,
                      floating[::1] eta,
                      floating[::1] q,
                      floating[::1] eta_diff,
                      floating[::1] u_logs,
                      floating[::1] half_var_tau,
                      floating[::1] mu_mult,
                      floating dq_scale,
                      int threads,
                      bint use_blas,
                      bint low_memory) noexcept nogil


cpdef void cpp_e_step_mixture(int[::1] ld_left_bound,
                              integral[::1] ld_indptr,
                              noncomplex_numeric[::1] ld_data,
                              floating[::1] std_beta,
                              floating[:, ::1] var_gamma,
                              floating[:, ::1] var_mu,
                              floating[::1] eta,
                              floating[::1] q,
                              floating[::1] eta_diff,
                              floating[::1] log_null_pi,
                              floating[:, ::1] u_logs,
                              floating[:, ::1] half_var_tau,
                              floating[:, ::1] mu_mult,
                              floating dq_scale,
                              int threads,
                              bint use_blas,
                              bint low_memory) noexcept nogil

cpdef void cpp_e_step_grid(int[::1] ld_left_bound,
                           integral[::1] ld_indptr,
                           noncomplex_numeric[::1] ld_data,
                           floating[::1] std_beta,
                           floating[::1, :] var_gamma,
                           floating[::1, :] var_mu,
                           floating[::1, :] eta,
                           floating[::1, :] q,
                           floating[::1, :] eta_diff,
                           floating[::1, :] u_logs,
                           floating[::1, :] half_var_tau,
                           floating[::1, :] mu_mult,
                           floating dq_scale,
                           int[:] active_model_idx,
                           int threads,
                           bint use_blas,
                           bint low_memory) noexcept nogil



import numpy as np
import psutil


def fits_in_memory(alloc_size, max_prop=.9):
    """
    Check whether there's enough memory resources to load an object
    with the given allocation size (in MB).
    :param alloc_size: The allocation size
    :param max_prop: The maximum proportion of available memory allowed for the object
    """

    avail_mem = psutil.virtual_memory().available / (1024.0 ** 2)

    if alloc_size / avail_mem > max_prop:
        return False
    else:
        return True


def dict_concat(d, axis=0):
    """
    Concatenate the values of a dictionary into a single vector
    :param d: A dictionary where values are numeric scalars or vectors
    :param axis: Concatenate along given axis.
    """
    return np.concatenate([d[c] for c in sorted(d.keys())], axis=axis)


def dict_mean(d, axis=None):
    """
    Estimate the mean of the values of a dictionary
    :param d: A dictionary where values are numeric scalars or vectors
    :param axis: Perform aggregation along given axis.
    """
    return np.mean(np.array([np.mean(v, axis=axis) for v in d.values()]), axis=axis)


def dict_sum(d, axis=None, transform=None):
    """
    Estimate the sum of the values of a dictionary
    :param d: A dictionary where values are numeric scalars or vectors
    :param axis: Perform aggregation along given axis.
    :param transform: Transformation to apply before summing.
    """
    if transform is None:
        return np.sum(np.array([np.sum(v, axis=axis) for v in d.values()]), axis=axis)
    else:
        return np.sum(np.array([np.sum(transform(v), axis=axis) for v in d.values()]), axis=axis)


def dict_elementwise_transform(d, transform):
    """
    Apply a transformation to values of a dictionary
    :param d: A dictionary where values are numeric scalars or vectors
    :param transform: A function to apply to
    """
    return {c: np.vectorize(transform)(v) for c, v in d.items()}


def dict_elementwise_dot(d1, d2):
    """
    Apply element-wise product between the values of two dictionaries

    :param d1: A dictionary where values are numeric scalars or vectors
    :param d2: A dictionary where values are numeric scalars or vectors
    """
    return {c: d1[c]*d2[c] for c, v in d1.items()}


def dict_dot(d1, d2):
    """
    Perform dot product on the elements of d1 and d2
    :param d1: A dictionary where values are numeric scalars or vectors
    :param d2: A dictionary where values are numeric scalars or vectors
    """
    return np.sum([np.dot(d1[c], d2[c]) for c in d1.keys()])


def dict_set(d, value):
    """
    :param d: A dictionary where values are numeric vectors
    :param value: A value to set for all vectors
    """
    for c in d:
        d[c][:] = value

    return d


def dict_repeat(value, shapes):
    """
    Given a value, create a dictionary where the value is repeated
    according to the shapes parameter
    :param shapes: A dictionary of shapes. Key is arbitrary, value is integer input to np.repeat
    :param value:  The value to repeat
    """
    return {c: value*np.ones(shp) for c, shp in shapes.items()}


def expand_column_names(c_name, shape, sep='_'):
    """
    Given a desired column name `c_name` and a matrix `shape`
    that we'd like to apply the column name to, return a list of
    column names for every column in the matrix. The column names will be
    in the form of `c_name` followed by an index, separated by `sep`.

    For example, if the column name is `BETA`, the
    shape is (100, 3) and the separator is `_`, we return a list with:
    [`BETA_0`, `BETA_1`, `BETA_2`]

    If the matrix in question is a vector, we just return the column name
    without any indices appended to it.

    :param c_name: A string object
    :param shape: The shape of a numpy matrix or vector
    :param sep: The separator
    """

    if len(shape) < 2:
        return c_name
    elif shape[1] == 1:
        return c_name
    else:
        return [c_name + f'{sep}{i}' for i in range(shape[1])]



from cython cimport floating, integral

cpdef void e_step(int[::1] ld_left_bound,
                  integral[::1] ld_indptr,
                  floating[::1] ld_data,
                  floating[::1] std_beta,
                  floating[::1] var_gamma,
                  floating[::1] var_mu,
                  floating[::1] eta,
                  floating[::1] q,
                  floating[::1] eta_diff,
                  floating[::1] u_logs,
                  floating[::1] half_var_tau,
                  floating[::1] mu_mult,
                  int threads,
                  bint use_blas,
                  bint low_memory
                 ) noexcept nogil


cpdef void e_step_mixture(int[::1] ld_left_bound,
                          integral[::1] ld_indptr,
                          floating[::1] ld_data,
                          floating[::1] std_beta,
                          floating[:, ::1] var_gamma,
                          floating[:, ::1] var_mu,
                          floating[::1] eta,
                          floating[::1] q,
                          floating[::1] eta_diff,
                          floating[::1] log_null_pi,
                          floating[:, ::1] u_logs,
                          floating[:, ::1] half_var_tau,
                          floating[:, ::1] mu_mult,
                          int threads,
                          bint use_blas,
                          bint low_memory) noexcept nogil


cpdef void e_step_grid(int[::1] ld_left_bound,
                       integral[::1] ld_indptr,
                       floating[::1] ld_data,
                       floating[::1] std_beta,
                       floating[::1, :] var_gamma,
                       floating[::1, :] var_mu,
                       floating[::1, :] eta,
                       floating[::1, :] q,
                       floating[::1, :] eta_diff,
                       floating[::1, :] u_logs,
                       floating[::1, :] half_var_tau,
                       floating[::1, :] mu_mult,
                       int[:] active_model_idx,
                       int threads,
                       bint use_blas,
                       bint low_memory) noexcept nogil




from cython.parallel import prange, parallel
from ...utils.math_utils cimport (
    sigmoid,
    softmax,
    clip,
    scipy_blas_axpy,
    scipy_blas_dot,
    axpy,
    dot
)
from .e_step_cpp cimport cpp_blas_axpy, cpp_blas_dot
cimport cython
import numpy as np
cimport numpy as np
from cython cimport floating, integral


# A safe way to get the number of the thread currently executing the code:
# This is used to avoid compile-time errors when compiling the code with OpenMP support disabled.
# In earlier iterations, we used:
# cimport openmp
# openmp.omp_get_thread_num()
# But this tends to fail when OpenMP is not enabled.
# The code below is a safer way to get the thread number.
cdef extern from *:
    """
    #ifdef _OPENMP
    #include <omp.h>
    #else
    int omp_get_thread_num() { return 0; }
    #endif
    """
    int omp_get_thread_num() noexcept nogil


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
@cython.exceptval(check=False)
cpdef void update_q_factor_matrix(int[:] active_model_idx,
                                  int[::1] ld_left_bound,
                                  integral[::1] ld_indptr,
                                  floating[::1] ld_data,
                                  floating[::1, :] eta,
                                  floating[::1, :] q,
                                  bint use_blas,
                                  int threads) noexcept nogil:
    """
        Compute or update q factor, defined as the result of a dot product between the
        Linkage-disequilibrium matrix (LD) and the current value of eta (the posterior mean
        for the effect sizes). The definition of the q-factor excludes the diagonal elements.

        This function assumes that we have a matrix of eta values, where each column represents
        eta from a different model (this is relevant for grid-based models).

        The implementation below assumes that the eta and q matrices are column-major matrices.
    """

    cdef:
        int j, m, m_idx, start, end, c_size = eta.shape[0], num_models = active_model_idx.shape[0]
        integral ld_start, ld_end

    for j in prange(c_size, nogil=True, schedule='static', num_threads=threads):

        ld_start = ld_indptr[j]
        ld_end = ld_indptr[j + 1]
        start = ld_left_bound[j]
        end = start + (ld_end - ld_start)

        for m in range(num_models):
            m_idx = active_model_idx[m]
            if use_blas:
                q[j, m_idx] += cpp_blas_dot(eta[start:end, m_idx], ld_data[ld_start:ld_end])
            else:
                q[j, m_idx] += dot(eta[start:end, m_idx], ld_data[ld_start:ld_end])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
@cython.exceptval(check=False)
cpdef void update_q_factor(int[::1] ld_left_bound,
                           integral[::1] ld_indptr,
                           floating[::1] ld_data,
                           floating[::1] eta,
                           floating[::1] q,
                           bint use_blas,
                           int threads) noexcept nogil:

    """
        Compute or update q factor, defined as the result of a dot product between the
        Linkage-disequilibrium matrix (LD) and the current value of eta (the posterior mean
        for the effect sizes). The definition of the q-factor excludes the diagonal elements.

        This function assumes that we have a vector of eta values. If you have a matrix of eta
        values, you should use the `update_q_factor_matrix` function instead.
    """

    cdef:
        int j, start, end, c_size = eta.shape[0]
        integral ld_start, ld_end;

    for j in prange(c_size, nogil=True, schedule='static', num_threads=threads):

        ld_start = ld_indptr[j]
        ld_end = ld_indptr[j + 1]
        start = ld_left_bound[j]
        end = start + (ld_end - ld_start)

        if use_blas:
            q[j] += cpp_blas_dot(ld_data[ld_start:ld_end], eta[start:end])
        else:
            q[j] += dot(ld_data[ld_start:ld_end], eta[start:end])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
@cython.exceptval(check=False)
cpdef void e_step(int[::1] ld_left_bound,
                  integral[::1] ld_indptr,
                  floating[::1] ld_data,
                  floating[::1] std_beta,
                  floating[::1] var_gamma,
                  floating[::1] var_mu,
                  floating[::1] eta,
                  floating[::1] q,
                  floating[::1] eta_diff,
                  floating[::1] u_logs,
                  floating[::1] half_var_tau,
                  floating[::1] mu_mult,
                  int threads,
                  bint use_blas,
                  bint low_memory) noexcept nogil:

    """
    
    This function performs the E-step of the variational inference algorithm. It updates the variational parameters
    for the latent variables, and the posterior mean of the effect sizes. The function is written in Cython, and 
    compiled to C code. The function is called from the Python code in the vi function in the variational
    inference module (VIPRS.py).
    
    :param ld_left_bound: The index of the leftmost neighboring SNP in the LD matrix (i.e. index of the leftmost that is SNP in LD with SNP j).
    :param ld_indptr: The index pointers for the flattened LD matrix (`ld_data`).
    :param ld_data: The flattened LD matrix.
    :param std_beta: The standardized effect sizes.
    :param var_gamma: The variational gamma parameters.
    :param var_mu: The variational mu parameters.
    :param eta: The posterior mean of the effect sizes.
    :param q: The q factors.
    :param u_logs: The logarithmic factors used to compute the variational gamma parameters.
    :param half_var_tau: The reciprocal sigma factors used to compute the variational gamma parameters.
    :param mu_mult: The multiplicative factor that is used to compute the variational mu parameters.
    :param threads: The number of threads to use for parallelization.
    """

    cdef:
        int j, start, end, c_size = var_mu.shape[0]
        long ld_start, ld_end
        floating u_j

    for j in prange(c_size, nogil=True, schedule='static', num_threads=threads):

        # The start and end coordinates for the flattened LD matrix:
        ld_start = ld_indptr[j]
        ld_end = ld_indptr[j + 1]

        # The start and end coordinates for the neighboring SNPs:
        start = ld_left_bound[j]
        end = start + (ld_end - ld_start)

        # Compute the variational mu beta:
        var_mu[j] = mu_mult[j] * (std_beta[j] - q[j])

        # Compute the variational gamma:
        u_j = u_logs[j] + half_var_tau[j] * var_mu[j] * var_mu[j]
        var_gamma[j] = sigmoid(u_j)

        # Compute the difference between the new and old values for the posterior mean:
        eta_diff[j] = var_gamma[j] * var_mu[j] - eta[j]

        # Update the q factors for all neighboring SNPs that are in LD with SNP j
        if use_blas:
            cpp_blas_axpy(q[start:end], ld_data[ld_start:ld_end], eta_diff[j])
        else:
            axpy(q[start:end], ld_data[ld_start:ld_end], eta_diff[j])

        if start <= j < end:
            # Updating q factors for neighboring SNPs above also updates the q-factor for
            # SNP j, so we correct that here:
            q[j] = q[j] - eta_diff[j]

        # Update the posterior mean:
        eta[j] = eta[j] + eta_diff[j]

    if low_memory:
        # If the LD matrix used in the above operations is upper-triangular, then we would not
        # have updated the q-factors of variants based on the new etas for variants that come after them
        # in the dataset. So, we need to correct for that here:
        update_q_factor(ld_left_bound, ld_indptr, ld_data, eta_diff, q, use_blas, threads)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
@cython.exceptval(check=False)
cpdef void e_step_mixture(int[::1] ld_left_bound,
                          integral[::1] ld_indptr,
                          floating[::1] ld_data,
                          floating[::1] std_beta,
                          floating[:, ::1] var_gamma,
                          floating[:, ::1] var_mu,
                          floating[::1] eta,
                          floating[::1] q,
                          floating[::1] eta_diff,
                          floating[::1] log_null_pi,
                          floating[:, ::1] u_logs,
                          floating[:, ::1] half_var_tau,
                          floating[:, ::1] mu_mult,
                          int threads,
                          bint use_blas,
                          bint low_memory) noexcept nogil:

    """

    This function performs the E-step of the variational inference algorithm. It updates the variational parameters
    for the latent variables, and the posterior mean of the effect sizes. The function is written in Cython, and 
    compiled to C code. The function is called from the Python code in the vi function in the variational
    inference module (VIPRS.py).

    :param ld_left_bound: The index of the leftmost neighboring SNP in the LD matrix 
    (i.e. index of the leftmost that is SNP in LD with SNP j).
    :param ld_indptr: The index pointers for the flattened LD matrix (`ld_data`).
    :param ld_data: The flattened LD matrix.
    :param std_beta: The standardized effect sizes.
    :param var_gamma: The variational gamma parameters.
    :param var_mu: The variational mu parameters.
    :param eta: The posterior mean of the effect sizes.
    :param q: The q factors.
    :param u_logs: The logarithmic factors used to compute the variational gamma parameters.
    :param half_var_tau: The reciprocal sigma factors used to compute the variational gamma parameters.
    :param mu_mult: The multiplicative factor that is used to compute the variational mu parameters.
    :param threads: The number of threads to use for parallelization.
    """

    cdef:
        int j, start, end, thread_offset, c_size = var_mu.shape[0], k, K = var_mu.shape[1]
        int u_size = threads * (K + 1)
        long ld_start, ld_end
        floating mu_beta_j
        floating[::1] u_j

    with gil:
        u_j = np.empty(threads*(K + 1), dtype=[np.float64, np.float32][floating is float])

    for j in prange(c_size, nogil=True, schedule='static', num_threads=threads):

        # Set the thread offset for the u_j array:
        thread_offset = omp_get_thread_num() * (K + 1)

        # The start and end coordinates for the flattened LD matrix:
        ld_start = ld_indptr[j]
        ld_end = ld_indptr[j + 1]

        # The start and end coordinates for the neighboring SNPs:
        start = ld_left_bound[j]
        end = start + (ld_end - ld_start)

        mu_beta_j = std_beta[j] - q[j]

        for k in range(K):
            # Compute the variational mu beta:
            var_mu[j, k] = mu_mult[j, k] * mu_beta_j
            u_j[thread_offset + k] = u_logs[j, k] + half_var_tau[j, k] * var_mu[j, k] * var_mu[j, k]

        # Normalize the variational gammas:
        u_j[thread_offset + K] = log_null_pi[j]
        var_gamma[j, :] = softmax(u_j[thread_offset:thread_offset + K + 1])[:K]

        # Compute the difference between the new and old values for the posterior mean:
        eta_diff[j] = -eta[j]

        for k in range(K):
            eta_diff[j] += var_gamma[j, k] * var_mu[j, k]

        # Update the q factors for all neighboring SNPs that are in LD with SNP j
        if use_blas:
            cpp_blas_axpy(q[start:end], ld_data[ld_start:ld_end], eta_diff[j])
        else:
            axpy(q[start:end], ld_data[ld_start:ld_end], eta_diff[j])

        if start <= j < end:
            # Updating q factors for neighboring SNPs above also updates the q-factor for
            # SNP j, so we correct that here:
            q[j] = q[j] - eta_diff[j]

        # Update the posterior mean:
        eta[j] = eta[j] + eta_diff[j]

    if low_memory:
        # If the LD matrix used in the above operations is upper-triangular, then we would not
        # have updated the q-factors of variants based on the new etas for variants that come after them
        # in the dataset. So, we need to correct for that here:
        update_q_factor(ld_left_bound, ld_indptr, ld_data, eta_diff, q, use_blas, threads)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
@cython.cdivision(True)
@cython.exceptval(check=False)
cpdef void e_step_grid(int[::1] ld_left_bound,
                       integral[::1] ld_indptr,
                       floating[::1] ld_data,
                       floating[::1] std_beta,
                       floating[::1, :] var_gamma,
                       floating[::1, :] var_mu,
                       floating[::1, :] eta,
                       floating[::1, :] q,
                       floating[::1, :] eta_diff,
                       floating[::1, :] u_logs,
                       floating[::1, :] half_var_tau,
                       floating[::1, :] mu_mult,
                       int[:] active_model_idx,
                       int threads,
                       bint use_blas,
                       bint low_memory) noexcept nogil:

    """

    This function performs the E-step of the variational inference algorithm. It updates the variational parameters
    for the latent variables, and the posterior mean of the effect sizes. The function is written in Cython, and 
    compiled to C code. The function is called from the Python code in the vi function in the variational
    inference module (VIPRS.py).

    :param ld_left_bound: The index of the leftmost neighboring SNP in the LD matrix 
    (i.e. index of the leftmost that is SNP in LD with SNP j).
    :param ld_indptr: The index pointers for the flattened LD matrix (`ld_data`).
    :param ld_data: The flattened LD matrix.
    :param std_beta: The standardized effect sizes.
    :param var_gamma: The variational gamma parameters.
    :param var_mu: The variational mu parameters.
    :param eta: The posterior mean of the effect sizes.
    :param q: The q factors.
    :param u_logs: The logarithmic factors used to compute the variational gamma parameters.
    :param half_var_tau: The reciprocal sigma factors used to compute the variational gamma parameters.
    :param mu_mult: The multiplicative factor that is used to compute the variational mu parameters.
    :param threads: The number of threads to use for parallelization.
    """

    cdef:
        int start, end, j, c_size = var_mu.shape[0], m_idx, m, num_models = active_model_idx.shape[0]
        long ld_start, ld_end
        floating u_j

    for j in prange(c_size, nogil=True, schedule='static', num_threads=threads):

        # The start and end coordinates for the flattened LD matrix:
        ld_start = ld_indptr[j]
        ld_end = ld_indptr[j + 1]

        # The start and end coordinates for the neighboring SNPs:
        start = ld_left_bound[j]
        end = start + (ld_end - ld_start)

        for m_idx in range(num_models):

            # Retrieve the index of the active model:
            m = active_model_idx[m_idx]

            # Compute the variational mu beta:
            var_mu[j, m] = mu_mult[j, m] * (std_beta[j] - q[j, m])

            # Compute the variational gamma:
            u_j = u_logs[j, m] + half_var_tau[j, m] * var_mu[j, m] * var_mu[j, m]
            var_gamma[j, m] = sigmoid(u_j)

            # Compute the difference between the new and old values for the posterior mean:
            eta_diff[j, m] = var_gamma[j, m] * var_mu[j, m] - eta[j, m]

            # Update the q factors for all neighboring SNPs that are in LD with SNP j
            if use_blas:
                cpp_blas_axpy(q[start:end, m], ld_data[ld_start:ld_end], eta_diff[j, m])
            else:
                axpy(q[start:end, m], ld_data[ld_start:ld_end], eta_diff[j, m])

            if start <= j < end:
                # Updating q factors for neighboring SNPs above also updates the q-factor for
                # SNP j, so we correct that here:
                q[j, m] = q[j, m] - eta_diff[j, m]

            # Update the posterior mean:
            eta[j, m] = eta[j, m] + eta_diff[j, m]

    if low_memory:
        # If the LD matrix used in the above operations is upper-triangular, then we would not
        # have updated the q-factors of variants based on the new etas for variants that come after them
        # in the dataset. So, we need to correct for that here:
        update_q_factor_matrix(active_model_idx, ld_left_bound, ld_indptr, ld_data, eta_diff, q, use_blas, threads)




import magenpy as mgp
import viprs as vp
from viprs.model.vi.e_step_cpp import check_blas_support, check_omp_support
import numpy as np
from viprs.model.gridsearch.HyperparameterGrid import HyperparameterGrid
from viprs.model.gridsearch.VIPRSGrid import VIPRSGrid
import shutil
import pytest


@pytest.fixture(scope='module')
def gdl_object():
    """
    Initialize a GWADataLoader using data pre-packaged with magenpy.
    Make this data loader available to all tests.
    """
    gdl = mgp.GWADataLoader(mgp.tgp_eur_data_path(),
                            sumstats_files=mgp.ukb_height_sumstats_path(),
                            sumstats_format='fastgwa',
                            backend='xarray')

    # Compute LD matrix:
    gdl.compute_ld('shrinkage',
                   genetic_map_ne=11400,
                   genetic_map_sample_size=183,
                   output_dir='output/ld/')

    yield gdl

    # Clean up after tests are done:
    gdl.cleanup()
    shutil.rmtree(gdl.temp_dir)
    shutil.rmtree(gdl.output_dir)


@pytest.fixture(scope='module')
def viprs_model(gdl_object):
    """
    Initialize a VIPRS model using data pre-packaged with magenpy.
    Make this data loader available to all tests.
    """
    return vp.VIPRS(gdl_object, verbose=False)


@pytest.fixture(scope='module')
def viprsmix_model(gdl_object):
    """
    Initialize a VIPRS model using data pre-packaged with magenpy.
    Make this data loader available to all tests.
    """
    return vp.VIPRSMix(gdl_object, K=10, verbose=False)


@pytest.fixture(scope='module')
def viprsgrid_model(gdl_object):
    """
    Initialize a VIPRS model using data pre-packaged with magenpy.
    Make this data loader available to all tests.
    """
    # Create a grid:
    grid = HyperparameterGrid()
    # Generate a grid for pi using 5 equidistant grid points:
    grid.generate_pi_grid(steps=5)
    # Generate a grid for sigma epsilon using 5 equidistant grid points:
    grid.generate_sigma_epsilon_grid(steps=5)

    return VIPRSGrid(gdl_object, grid, verbose=False)


class TestVIPRS(object):

    def test_viprs_init(self, viprs_model, gdl_object):

        assert viprs_model.m == gdl_object.m

        viprs_model.initialize()

        assert 0. < viprs_model.pi < 1.
        assert 0. < viprs_model.sigma_epsilon < 1.
        assert viprs_model.tau_beta > 0.

        for p in (viprs_model.var_gamma, viprs_model.var_mu, viprs_model.var_tau,
                  viprs_model.q, viprs_model.eta, viprs_model.zeta):
            assert p[22].shape == (viprs_model.m, )
            assert p[22].dtype == viprs_model.float_precision

        # ...

    def test_viprs_fit(self, viprs_model):

        viprs_model.fit(max_iter=5)

        for p in (viprs_model.pip, viprs_model.post_mean_beta, viprs_model.post_var_beta):
            assert p[22].shape == (viprs_model.m,)
            assert p[22].dtype == viprs_model.float_precision


class TestVIPRSMix(TestVIPRS):

    def test_viprs_init(self, viprsmix_model, gdl_object):

        assert viprsmix_model.m == gdl_object.m

        viprsmix_model.initialize()

        assert np.all((0. < viprsmix_model.pi) & (viprsmix_model.pi < 1.))
        assert 0. < np.sum(viprsmix_model.pi) < 1.
        assert np.all((0. < viprsmix_model.sigma_epsilon) & (viprsmix_model.sigma_epsilon < 1.))
        assert np.all(viprsmix_model.tau_beta > 0.)

        for p in (viprsmix_model.var_gamma, viprsmix_model.var_mu, viprsmix_model.var_tau):
            assert p[22].shape == (viprsmix_model.m, 10)
            assert p[22].dtype == viprsmix_model.float_precision

        for p in (viprsmix_model.q, viprsmix_model.eta, viprsmix_model.zeta):
            assert p[22].shape == (viprsmix_model.m, )
            assert p[22].dtype == viprsmix_model.float_precision


class TestVIPRSGrid(TestVIPRS):

    def test_viprs_init(self, viprsgrid_model, gdl_object):

        assert viprsgrid_model.m == gdl_object.m

        viprsgrid_model.initialize()

        assert np.all((0. < viprsgrid_model.pi) & (viprsgrid_model.pi < 1.))
        assert np.all((0. < viprsgrid_model.sigma_epsilon) & (viprsgrid_model.sigma_epsilon < 1.))
        assert np.all(viprsgrid_model.tau_beta > 0.)

        for p in (viprsgrid_model.var_gamma, viprsgrid_model.var_mu, viprsgrid_model.var_tau,
                  viprsgrid_model.q, viprsgrid_model.eta, viprsgrid_model.zeta):
            assert p[22].shape == (viprsgrid_model.m, 25)
            assert p[22].dtype == viprsgrid_model.float_precision

        # ...

    def test_viprs_fit(self, viprsgrid_model):

        viprsgrid_model.fit(max_iter=5)

        for p in (viprsgrid_model.pip, viprsgrid_model.post_mean_beta, viprsgrid_model.post_var_beta):
            assert p[22].shape == (viprsgrid_model.m, 25)
            assert p[22].dtype == viprsgrid_model.float_precision


@pytest.mark.xfail(not check_blas_support(), reason="BLAS library not found!")
def test_check_blas_support():
    assert check_blas_support()


@pytest.mark.xfail(not check_omp_support(), reason="OpenMP library not found!")
def test_check_omp_support():
    assert check_omp_support()



import numpy as np
import pandas as pd
import itertools


class HyperparameterGrid(object):
    """
    A utility class to facilitate generating grids for the
    hyperparameters of the standard `VIPRS` models. It is designed to
    interface with models that operate on grids of hyperparameters,
    such as `VIPRSGridSeach` and `VIPRSBMA`. The hyperparameters for
    the standard VIPRS model are:

    * `sigma_epsilon`: The residual variance for the phenotype.
    * `tau_beta`: The precision (inverse variance) of the prior for the effect sizes.
    * `pi`: The proportion of non-zero effect sizes (polygenicity).

    :ivar sigma_epsilon: A grid of values for the residual variance hyperparameter.
    :ivar tau_beta: A grid of values for the precision of the prior for the effect sizes.
    :ivar pi: A grid of values for the proportion of non-zero effect sizes.
    :ivar h2_est: An estimate of the heritability for the trait under consideration.
    :ivar h2_se: The standard error of the heritability estimate.
    :ivar n_snps: The number of common variants that may be relevant for this analysis.

    """

    def __init__(self,
                 sigma_epsilon_grid=None,
                 sigma_epsilon_steps=None,
                 tau_beta_grid=None,
                 tau_beta_steps=None,
                 pi_grid=None,
                 pi_steps=None,
                 h2_est=None,
                 h2_se=None,
                 n_snps=1e6):
        """

        Create a hyperparameter grid for the standard VIPRS model with the
        spike-and-slab prior. The hyperparameters for this model are:

        * `sigma_epsilon`: The residual variance
        * `tau_beta`: The precision (inverse variance) of the prior for the effect sizes
        * `pi`: The proportion of non-zero effect sizes

        For each of these hyperparameters, we can provide a grid of values to search over.
        If the heritability estimate and standard error (from e.g. LDSC) are provided,
        we can generate grids for sigma_epsilon and tau_beta that are informed by these estimates.

        For each hyperparameter to be included in the grid, user must specify either the grid
        itself, or the number of steps to use to generate the grid.

        :param sigma_epsilon_grid: An array containing a grid of values for the sigma_epsilon hyperparameter.
        :param sigma_epsilon_steps: The number of steps for the sigma_epsilon grid
        :param tau_beta_grid: An array containing a grid of values for the tau_beta hyperparameter.
        :param tau_beta_steps: The number of steps for the tau_beta grid
        :param pi_grid: An array containing a grid of values for the pi hyperparameter
        :param pi_steps: The number of steps for the pi grid
        :param h2_est: An estimate of the heritability for the trait under consideration. If provided,
        we can generate grids for some of the hyperparameters that are consistent with this estimate.
        :param h2_se: The standard error of the heritability estimate. If provided, we can generate grids
        for some of the hyperparameters that are consistent with this estimate.
        :param n_snps: Number of common variants that may be relevant for this analysis. This estimate can
        be used to generate grids that are based on this number.
        """

        # If the heritability estimate is not provided, use a reasonable default value of 0.1
        # with a wide standard error of 0.1.
        if h2_est is None:
            self.h2_est = 0.1
            self.h2_se = 0.1
        else:
            self.h2_est = h2_est
            self.h2_se = h2_se

        self.n_snps = n_snps
        self._search_params = []

        # Initialize the grid for sigma_epsilon:
        self.sigma_epsilon = sigma_epsilon_grid
        if self.sigma_epsilon is not None:
            self._search_params.append('sigma_epsilon')
        elif sigma_epsilon_steps is not None:
            self.generate_sigma_epsilon_grid(steps=sigma_epsilon_steps)

        # Initialize the grid for the tau_beta:
        self.tau_beta = tau_beta_grid
        if self.tau_beta is not None:
            self._search_params.append('tau_beta')
        elif tau_beta_steps is not None:
            self.generate_tau_beta_grid(steps=tau_beta_steps)

        # Initialize the grid for pi:
        self.pi = pi_grid
        if self.pi is not None:
            self._search_params.append('pi')
        elif pi_steps is not None:
            self.generate_pi_grid(steps=pi_steps)

    def _generate_h2_grid(self, steps=5):
        """
        Use the heritability estimate and standard error to generate a grid of values for
        the heritability parameter. Specifically, given the estimate and standard error, we
        generate heritability estimates from the percentiles of the normal distribution,
        with mean `h2_est` and standard deviation `h2_se`. The grid values range from the 10th
        percentile to the 90th percentile of this normal distribution.

        :param steps: The number of steps for the heritability grid.
        :return: A grid of values for the heritability parameter.

        """

        assert steps > 0
        assert self.h2_est is not None

        # If the heritability standard error is not provided, we use half of the heritability estimate
        # by default.
        # *Justification*: Under the assumption that heritability for the trait being analyzed
        # is significantly greater than 0, the standard error should be, at a maximum,
        # half of the heritability estimate itself to get us a Z-score with absolute value
        # greater than 2.
        if self.h2_se is None:
            h2_se = self.h2_est * 0.5
        else:
            h2_se = self.h2_se

        # Sanity checking steps:
        assert 0. < self.h2_est < 1.
        assert h2_se > 0

        from scipy.stats import norm

        # First, determine the percentile boundaries to avoid producing
        # invalid values for the heritability grid:

        percentile_start = max(0.1, norm.cdf(1e-5, loc=self.h2_est, scale=h2_se))
        percentile_stop = min(0.9, norm.cdf(1. - 1e-5, loc=self.h2_est, scale=h2_se))

        # Generate the heritability grid:
        return norm.ppf(np.linspace(percentile_start, percentile_stop, steps),
                        loc=self.h2_est, scale=h2_se)

    def generate_sigma_epsilon_grid(self, steps=5):
        """
        Generate a grid of values for the `sigma_epsilon` (residual variance) hyperparameter.

        :param steps: The number of steps for the sigma_epsilon grid.
        """

        assert steps > 0

        h2_grid = self._generate_h2_grid(steps)
        self.sigma_epsilon = 1. - h2_grid

        if 'sigma_epsilon' not in self._search_params:
            self._search_params.append('sigma_epsilon')

    def generate_tau_beta_grid(self, steps=5):
        """
        Generate a grid of values for the `tau_beta`
        (precision of the prior for the effect sizes) hyperparameter.
        :param steps: The number of steps for the `tau_beta` grid
        """

        assert steps > 0

        h2_grid = self._generate_h2_grid(steps)
        # Assume ~1% of SNPs are causal:
        self.tau_beta = 0.01*self.n_snps / h2_grid

        if 'tau_beta' not in self._search_params:
            self._search_params.append('tau_beta')

    def generate_pi_grid(self, steps=5):
        """
        Generate a grid of values for the `pi` (proportion of non-zero effect sizes) hyperparameter.
        :param steps: The number of steps for the `pi` grid
        """

        assert steps > 0

        self.pi = np.unique(np.clip(10. ** (-np.linspace(np.floor(np.log10(self.n_snps)), 0., steps)),
                                    a_min=1. / self.n_snps,
                                    a_max=1. - 1. / self.n_snps))

        if 'pi' not in self._search_params:
            self._search_params.append('pi')

    def combine_grids(self):
        """
        Weave together the different hyperparameter grids and return a list of
        dictionaries where the key is the hyperparameter name and the value is
        value for that hyperparameter.

        :return: A list of dictionaries containing the hyperparameter values.
        :raises ValueError: If all the grids are empty.

        """
        hyp_names = [name for name, value in self.__dict__.items()
                     if value is not None and name in self._search_params]

        if len(hyp_names) > 0:
            hyp_values = itertools.product(*[hyp_grid for hyp_name, hyp_grid in self.__dict__.items()
                                             if hyp_grid is not None and hyp_name in hyp_names])

            return [dict(zip(hyp_names, hyp_v)) for hyp_v in hyp_values]
        else:
            raise ValueError("All the grids are empty!")

    def to_table(self):
        """
        :return: The hyperparameter grid as a pandas `DataFrame`.
        """

        combined_grids = self.combine_grids()
        if combined_grids:
            return pd.DataFrame(combined_grids)




class OptimizeResult(object):
    """
    A class to store the results/progress of an optimization algorithm.
    Similar to the `OptimizeResult` class from `scipy.optimize`,
    but with a few additional fields.
    """

    def __init__(self):

        self.message = None
        self.stop_iteration = None
        self.success = None
        self.fun = None
        self.nit = None

    @property
    def iterations(self):
        """
        Return the number of iterations at its current value.
        """
        return self.nit

    @property
    def objective(self):
        """
        Return the objective function value at its current value.
        """
        return self.fun

    def reset(self):
        """
        Reset the stored values to their initial state.
        """

        self.message = None
        self.stop_iteration = False
        self.success = False
        self.fun = None
        self.nit = 0

    def update(self, fun, stop_iteration=False, success=False, message=None):
        """
        Update the stored values with new values.
        :param fun: The new objective function value
        :param stop_iteration: A flag to indicate whether the optimization algorithm has stopped iterating
        :param success: A flag to indicate whether the optimization algorithm has succeeded
        :param message: A detailed message about the optimization result.
        """

        self.fun = fun
        self.stop_iteration = stop_iteration
        self.success = success
        self.message = message

        self.nit += 1

    def __str__(self):
        return str(self.__dict__)



from setuptools import setup, Extension, find_packages
from extension_helpers import add_openmp_flags_if_available
from extension_helpers._openmp_helpers import check_openmp_support
import pkgconfig
import numpy as np
import warnings
import os


try:
    from Cython.Build import cythonize
except ImportError:
    warnings.warn("Cython not found.")
    cythonize = None

# ------------------------------------------------------
# Find and set BLAS-related flags and paths:


def find_blas_libraries():
    """
    Find BLAS libraries on the system using pkg-config.
    This function will return the include directories (compiler flags)
    and the linker flags to enable building the C/C++/Cython extensions
    that require BLAS (or whose performance would be enhanced with BLAS).

    We use pkg-config (as encapsulated in the `pkgconfig` Python package)
    to perform this search. Note that we augment the pkg-config
    search path with the conda library path (if available) to
    enable linking against BLAS libraries installed via Conda.

    :return: A dictionary with the following keys:
        * 'found': A boolean indicating whether BLAS libraries were found.
        * 'include_dirs': A list of include directories (compiler flags).
        * 'extra_link_args': A list of linker flags.
        * 'define_macros': A list of macros to define.
        * 'libraries': A list of libraries to link against.
    """

    # STEP 0: Get the current pkg-config search path:
    current_pkg_config_path = os.getenv("PKG_CONFIG_PATH", "")

    # STEP 1: Augment the pkg-config search path with
    # the path of the current Conda environment (if exists).
    # This can leverage BLAS libraries installed via Conda.

    conda_path = os.getenv("CONDA_PREFIX")

    if conda_path is not None:
        conda_pkgconfig_path = os.path.join(conda_path, 'lib/pkgconfig')
        if os.path.isdir(conda_pkgconfig_path):
            current_pkg_config_path += ":" + conda_pkgconfig_path

    # STEP 2: Add the updated path to the environment variable:
    os.environ["PKG_CONFIG_PATH"] = current_pkg_config_path

    # STEP 3: Get all pkg-config packages and filter to
    # those that have "blas" in the name.
    blas_packages = [pkg for pkg in pkgconfig.list_all()
                     if "blas" in pkg]

    # First check: Make sure that compiler flags are defined and a
    # valid cblas.h header file exists in the include directory:
    if len(blas_packages) >= 1:

        blas_packages = [pkg for pkg in blas_packages
                         if pkgconfig.cflags(pkg) and
                         os.path.isfile(os.path.join(pkgconfig.variables(pkg)['includedir'], 'cblas.h'))]

    # If there remains more than one library after the previous
    # search and filtering steps, then apply some heuristics
    # to select the most relevant one:
    if len(blas_packages) > 1:
        # Check if the information about the most relevant library
        # can be inferred from numpy. Note that this interface from
        # numpy changes quite often between versions, so it's not
        # a reliable check. But in case it works on some systems,
        # we use it to link to the same library as numpy:
        try:
            for pkg in blas_packages:
                if pkg in np.__config__.get_info('blas_opt')['libraries']:
                    blas_packages = [pkg]
                    break
        except (KeyError, AttributeError):
            pass

    # If there are still multiple libraries, then apply some
    # additional heuristics (based on name matching) to select
    # the most relevant one. Some libraries (e.g. flexiblas) are published with support for 64bit
    # and they expose libraries for non-BLAS API (with the _api suffix).
    # Ignore these here if that is the case?
    if len(blas_packages) > 1:
        # Some libraries (e.g. flexiblas) are published with support for 64bit
        # and they expose libraries for non-BLAS API (with the _api suffix).
        # Ignore these here if that is the case?

        idx_to_remove = set()

        for pkg1 in blas_packages:
            if pkg1 != 'blas':
                for i, pkg2 in enumerate(blas_packages):
                    if pkg1 != pkg2 and pkg1 in pkg2:
                        idx_to_remove.add(i)

        blas_packages = [pkg for i, pkg in enumerate(blas_packages) if i not in idx_to_remove]

    # After applying all the heuristics, out of all the remaining libraries,
    # select the first one in the list. Not the greatest solution, maybe
    # down the line we can use the same BLAS order as numpy.
    if len(blas_packages) >= 1:
        final_blas_pkg = blas_packages[0]
    else:
        final_blas_pkg = None

    # STEP 4: If a relevant BLAS package was found, extract the flags
    # needed for building the Cython/C/C++ extensions:

    if final_blas_pkg is not None:
        blas_info = pkgconfig.parse(final_blas_pkg)
        blas_info['define_macros'] = [('HAVE_CBLAS', None)]
    else:
        blas_info = {
            'include_dirs': [],
            'library_dirs': [],
            'libraries': [],
            'define_macros': [],
        }
        warnings.warn("""
            ********************* WARNING *********************
            BLAS library header files not found on your system. 
            This may slow down some computations. If you are 
            using conda, we recommend installing BLAS libraries 
            beforehand.
            ********************* WARNING *********************
        """, stacklevel=2)

    return blas_info


blas_flags = find_blas_libraries()

# ------------------------------------------------------
# Build cython extensions:


def no_cythonize(cy_extensions, **_ignore):
    """
    Copied from:
    https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html#distributing-cython-modules
    """

    for ext in cy_extensions:
        sources = []
        for s_file in ext.sources:
            path, ext = os.path.splitext(s_file)
            if ext in (".pyx", ".py"):
                if ext.language == "c++":
                    ext = ".cpp"
                else:
                    ext = ".c"
                s_file = path + ext
            sources.append(s_file)
        ext.sources[:] = sources

    return extensions


extensions = [
    Extension("viprs.utils.math_utils",
              ["viprs/utils/math_utils.pyx"],
              libraries=[[], ["m"]][os.name != 'nt'],  # Only include for non-Windows systems
              include_dirs=[np.get_include()],
              extra_compile_args=["-O3"]),
    Extension("viprs.model.vi.e_step",
              ["viprs/model/vi/e_step.pyx"],
              include_dirs=[np.get_include()],
              define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
              extra_compile_args=["-O3"]),
    Extension("viprs.model.vi.e_step_cpp",
              ["viprs/model/vi/e_step_cpp.pyx"],
              language="c++",
              libraries=blas_flags['libraries'],
              include_dirs=[np.get_include()] + blas_flags['include_dirs'],
              library_dirs=blas_flags['library_dirs'],
              define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")] + blas_flags['define_macros'],
              extra_compile_args=["-O3", "-std=c++17"])
]

if check_openmp_support():
    # Add any extension that requires openMP here:
    openmp_extensions = ['viprs.model.vi.e_step_cpp', 'viprs.model.vi.e_step']

    for omp_ext in extensions:
        if omp_ext.name in openmp_extensions:
            add_openmp_flags_if_available(omp_ext)
else:
    warnings.warn("""
        ******************** WARNING ********************
        OpenMP library not found on your system. This 
        means that some computations may be slower than 
        expected. It will preclude using multithreading 
        in the coordinate ascent optimization algorithm.
        ******************** WARNING ********************
    """)


if cythonize is not None:
    compiler_directives = {
        "language_level": 3,
        "embedsignature": True,
        'boundscheck': False,
        'wraparound': False,
        'nonecheck': False,
        'cdivision': True
    }
    extensions = cythonize(extensions, compiler_directives=compiler_directives)
else:
    extensions = no_cythonize(extensions)

# ------------------------------------------------------
# Read description/dependencies from file:

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt") as fp:
    install_requires = fp.read().strip().split("\n")

with open("requirements-optional.txt") as fp:
    opt_requires = fp.read().strip().split("\n")

with open("requirements-test.txt") as fp:
    test_requires = fp.read().strip().split("\n")

with open("requirements-docs.txt") as fp:
    doc_requires = fp.read().strip().split("\n")

# ------------------------------------------------------

setup(
    name="viprs",
    version="0.1.1",
    author="Shadi Zabad",
    author_email="shadi.zabad@mail.mcgill.ca",
    description="Variational Inference of Polygenic Risk Scores (VIPRS)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/shz9/viprs",
    classifiers=[
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12'
    ],
    package_dir={'': '.'},
    packages=find_packages(),
    python_requires=">=3.8",
    package_data={'viprs': ['model/vi/*.pxd', 'utils/*.pxd']},
    scripts=['bin/viprs_fit', 'bin/viprs_score', 'bin/viprs_evaluate'],
    install_requires=install_requires,
    extras_require={'opt': opt_requires, 'test': test_requires, 'docs': doc_requires},
    ext_modules=extensions,
    zip_safe=False
)




from .VIPRSGrid import VIPRSGrid
import numpy as np
from scipy.special import softmax


class VIPRSBMA(VIPRSGrid):
    """
    The `VIPRSBMA` class is an extension of the `VIPRSGrid` class that
    implements Bayesian model averaging for the `VIPRS` models in the grid.
    Bayesian model averaging is a technique that allows us to combine the
    results of multiple models by weighting them according to their evidence.
    In this context, we weigh the model by their final ELBO values.

    For more details on the BMA procedure implemented here, refer to the
    Supplementary material of:

    > Zabad S, Gravel S, Li Y. Fast and accurate Bayesian polygenic risk modeling with variational inference.
    Am J Hum Genet. 2023 May 4;110(5):741-761. doi: 10.1016/j.ajhg.2023.03.009.
    Epub 2023 Apr 7. PMID: 37030289; PMCID: PMC10183379.

    """

    def __init__(self,
                 gdl,
                 grid,
                 **kwargs):
        """

        Initialize the `VIPRSBMA` model.

        :param gdl: An instance of `GWADataLoader`
        :param grid: An instance of `HyperparameterGrid`
        :param kwargs: Additional keyword arguments for the VIPRS model
        """

        super().__init__(gdl, grid=grid, **kwargs)

    def average_models(self, normalization='softmax'):
        """
        Use Bayesian model averaging (BMA` to obtain final weights for each parameter.
        We average the weights by using the final ELBO for each model.
        
        :param normalization: The normalization scheme for the final ELBOs. Options are (`softmax`, `sum`). 
        :raises KeyError: If the normalization scheme is not recognized.
        """

        elbos = self.history['ELBO'][-1]

        if normalization == 'softmax':
            weights = np.array(softmax(elbos))
        elif normalization == 'sum':
            weights = np.array(elbos)

            # Correction for negative ELBOs:
            weights = weights - weights.min() + 1.
            weights /= weights.sum()
        else:
            raise KeyError("Normalization scheme not recognized. Valid options are: `softmax`, `sum`. "
                           "Got: {}".format(normalization))

        if int(self.verbose) > 1:
            print("Averaging PRS models with weights:", weights)

        for param in (self.pip, self.post_mean_beta, self.post_var_beta,
                      self.var_gamma, self.var_mu, self.var_tau,
                      self.eta, self.zeta, self.q):

            for c in param:
                param[c] = (param[c]*weights).sum(axis=1)

        return self
